import multiprocessing
from urllib3 import connectionpool, poolmanager

# we're not able to make a configuration option for this at the moment.
# It's unlikely anyone would want to get past twice the numbers of CPU
# On the other hand, this potential waste is really minimal, compared to
# what we're otherwise doing (Selenium…)
MAX_CONN = multiprocessing.cpu_count() * 3


class HTTPConnectionPool(connectionpool.HTTPConnectionPool):
    def __init__(self, *args, **kwargs):
        # this block=True will avoid hard failures if the pool is full,
        # but there's still something that flips it back to False sometimes
        # in those tests.
        kwargs.update(maxsize=MAX_CONN, block=True)
        super(HTTPConnectionPool, self).__init__(*args, **kwargs)


class HTTPSConnectionPool(connectionpool.HTTPSConnectionPool):
    def __init__(self, *args, **kwargs):
        # this block=True will avoid hard failures if the pool is full,
        # but there's still something that flips it back to False sometimes
        # in those tests.
        kwargs.update(maxsize=MAX_CONN, block=True)
        super(HTTPSConnectionPool, self).__init__(*args, **kwargs)


registry = poolmanager.pool_classes_by_scheme
registry['http'] = HTTPConnectionPool
registry['https'] = HTTPSConnectionPool



# Copyright 2018 Paul Morelle <madprog@htkc.org>
# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import absolute_import
import docker
import sys
_client = docker.from_env()


def heptapod_exec(ct, command, user='root'):
    container = _client.containers.get(ct)
    exit_code, output = container.exec_run(command, tty=True, demux=True,
                                           user=user)
    sys.stdout.write('+ docker exec heptapod {command}\n'.format(
        command=command,
    ))
    out, err = [o.decode() if o is not None else None for o in output]
    if out is not None:
        sys.stdout.write(out)
    if err is not None:
        sys.stderr.write(err)

    return exit_code, out


def heptapod_run_shell(ct, command, **kw):
    exit_code, output = heptapod_exec(ct, command, **kw)
    if exit_code:
        raise RuntimeError(
            ('Heptapod command {command} returned a non-zero '
             'exit code {exit_code}').format(command=command,
                                             exit_code=exit_code,
                                             ))


def heptapod_put_archive(ct, dest, path):
    """Put the tar archive at path at given dest in container."""
    container = _client.containers.get(ct)
    with open(path, 'rb') as arf:
        container.put_archive(dest, arf.read())


def heptapod_put_archive_bin(ct, dest, fobj):
    """Put the tar archive of file-like `fobj` at given dest in container."""
    _client.containers.get(ct).put_archive(dest, fobj.read())


def heptapod_get_archive(ct, path, tarf):
    """Get the file or directory at path as a tar archive.

    The tar binary contents is written to the tarf file-like object and
    a stats dict is returned.

    :param tarf: a binary file-like object
    :param archive_path: path to retrieve inside the Heptapod container
    :returns: dict of stats.
    """
    container = _client.containers.get(ct)
    bits, stats = container.get_archive(path)
    for chunk in bits:
        tarf.write(chunk)
    return stats


def host_address(ct):
    """Return a suitable address for the host, as seen from the container.

    For now this is just the IPv4 gateway, assumed to be the host, i.e. where
    these tests run.

    It will probably fail for more complicated network settings, but that's
    good enough for the time being.
    """
    container = _client.containers.get(ct)
    return container.attrs['NetworkSettings']['Gateway']



# Copyright 2018 Paul Morelle <madprog@htkc.org>
# Copyright 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import attr
from base64 import b64decode
import hashlib
from io import BytesIO
import logging
import os
import py
import re
import requests
from urllib.parse import (
    quote as urlquote,
    urlparse,
)
import time
import warnings

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from .access_levels import ProjectAccess
from .api import (
    api_request,
    GitLabEntity,
)
from .hg import LocalRepo
from .namespace import UserNameSpace
from .merge_request import MergeRequest
from .project_mirror import ProjectMirror
from .selenium import (
    assert_webdriver_not_error,
    could_click_element,
    raw_page_content,
    wait_assert_in_page_source,
    wait_could_click_element,
    webdriver_wait,
)
from .wait import (
    BASE_TIMEOUT,
    wait_assert,
)

logger = logging.getLogger(__name__)


def extract_gitlab_branch_titles(branches):
    return {name: info['commit']['title']
            for name, info in branches.items()}


def branch_title(all_branches, branch_name):
    """Return the title of head commit for the branch with given name.

    If no branch has the given name, ``None`` is returned.
    This is useful for one-liners (lambda) in ``wait_assert`` methods.
    """
    branch = all_branches.get(branch_name)
    if branch is None:
        return None

    return branch['commit']['title']


def protected_branch_api_subpath(spec=None):
    path = ['protected_branches']
    if spec is not None:
        path.append(urlquote(spec, safe=''))
    return '/'.join(path)


@attr.s
class Project(GitLabEntity):
    heptapod = attr.ib()
    name = attr.ib()
    group = attr.ib()
    # owner=None or id=None means it's not known yet, this should happen only
    # in the present module functions
    owner = attr.ib(default=None)
    id = attr.ib(default=None)
    # If vcs_type is None, this means Mercurial.
    # These functional tests should only exceptionally make a difference
    # between various ways Mercurial is supported ('hg_git', 'hgitaly'), and
    # preferably for temporary workarounds.
    vcs_type = attr.ib(default=None)
    is_legacy = attr.ib(default=False)
    hg_git_repo_expected = attr.ib(default=False)
    """For Mercurial, express whether an auxiliary Git repo is expected.

    This can be ``True`` in two cases:

    - because of general testing options. As of this writing, if
      ``--heptapod-hg-native`` is not set to ``without-git``.
    - because the Git repository is a functionally desireable outcome
      (see heptapod#125). In this case, the tests are expected to override
      the value, because it is initialized from the general testing options.
    """

    def owner_get(self, **kwargs):
        """A shortcut to perform a simple GET, with BasicAuth as the owner.

        All `kwargs` are passed directly to `requests.get()`
        """
        return requests.get(self.url, auth=self.owner_credentials, **kwargs)

    def get_session_cookie(self, webdriver):
        for cookie in webdriver.get_cookies():
            if cookie['name'].startswith('_gitlab_session'):
                return cookie
        raise LookupError("Could not find GitLab session cookie")

    def session_api_get(self, webdriver, subpath='', **kwargs):
        """Perform a simple GET, with the session cookie found in webdriver.

        The full URL is made of the API URL of the project, together with
        the given subpath (example '/merge_requests/1')
        """
        cookie = self.get_session_cookie(webdriver)
        return requests.get(
            '/'.join((self.api_url, subpath)),
            cookies={cookie['name']: cookie['value']})

    def api_edit(self, **params):
        """Perform a project API edit."""
        return api_request('PUT', self, self.owner_user, data=params)

    @property
    def owner_user(self):
        return self.heptapod.get_user(self.owner)

    def api_hgrc_set(self, user=None, **values):
        return self.user_api_request('PUT', user=user,
                                     subpath='hgrc', data=values)

    def owner_api_hgrc_get(self):
        return self.api_get(subpath='hgrc')

    def owner_api_hgrc_reset(self):
        return self.api_delete(subpath='hgrc')

    def owner_api_hg_heptapod_config(self):
        return self.api_get(subpath='hg_heptapod_config')

    def api_request(self, method, user, subpath='', **kwargs):
        """Perform a simple API HTTP request as the given user.

        `method` is the HTTP method to use, same as in `requests.request`.

        The full URL is made of the API URL of the project, together with
        the given subpath (example 'merge_requests/1').

        Appropriate authentication headers are added on the fly.

        All kwargs are passed to `requests.request()`
        """
        warnings.warn("Project.api_request() is deprecated. "
                      "Use user_api_request() instead.",
                      DeprecationWarning, stacklevel=2)
        return self.user_api_request(method, user=user, subpath=subpath,
                                     **kwargs)

    def api_get_field(self, key):
        """Return the value of a field by performing an API GET request.

        The request is made with full owner credentials.
        """
        resp = self.api_get()
        assert resp.status_code < 400
        return resp.json().get(key)

    def api_get_info(self):
        """Return the full project information from API.

        The request is made with full owner credentials.
        """
        resp = self.api_get()
        assert resp.status_code < 400
        return resp.json()

    def api_get_commit_metadata(self, gl_sha, check=True):
        """Retrieve commit metadata with the API.

        :param gl_sha: the primary identifier of the commit from
           GitLab's point of view, hence a Git SHA if VCS type is `hg_git`.

        - does not have file content nor any diff
        - works with obsolete commits, unless garbage collected.
        """
        resp = self.api_get(subpath='repository/commits/' + gl_sha)
        if not check:
            return resp

        assert resp.status_code == 200
        return resp.json()

    @property
    def hg_native(self):
        # TODO simple placeholder for now. Later on we can
        # set it at creation / search time
        return self.heptapod.hg_native

    @property
    def owner_credentials(self):
        """Return (user, password)."""
        user = self.owner_user
        return user.name, user.password

    @property
    def owner_token(self):
        return self.owner_user.token

    @property
    def owner_webdriver(self):
        return self.owner_user.webdriver

    @property
    def relative_uri(self):
        return '/'.join((self.group.full_path, self.name))

    @property
    def url(self):
        return '/'.join((self.heptapod.url, self.relative_uri))

    @property
    def owner_ssh_params(self):
        """See `ssh_params()`
        """
        return self.ssh_params(self.owner)

    def ssh_params(self, user_name):
        """Provide command and URL to perform SSH operations as `user_name`
        Example::

           ('ssh -i /tmp/id_rsa', 'git@localhost:root/test_project.hg')

        """
        heptapod = self.heptapod
        ext = '.git' if self.vcs_type == 'git' else '.hg'
        url = '/'.join((heptapod.ssh_url, self.relative_uri + ext))
        return self.heptapod.get_user(user_name).ssh_command, url

    def git_ssh_params(self, user_name):
        """Similar to ssh_params, tailored for Git.
        """
        heptapod = self.heptapod
        cmd = self.ssh_params(user_name)[0] + ' -p %d' % heptapod.ssh_port
        address = '{heptapod.ssh_user}@{heptapod.host}:{path}'.format(
            heptapod=heptapod,
            path=self.relative_uri + '.git',
        )
        return cmd, address

    def basic_auth_url(self, user_name, pwd=None):
        """Produce an URL suitable for basic authentication, hence hg CLI.

        :param pwd: if not supplied, will be read from the permanent users
                    known of :attr:`heptapod`
        """
        heptapod = self.heptapod
        if pwd is None:
            user = self.heptapod.get_user(user_name)
            pwd = user.token if user.vcs_token_only else user.password
        url = heptapod.parsed_url
        return "{scheme}://{auth}@{netloc}/{path}".format(
            scheme=url.scheme,
            netloc=url.netloc,
            auth=':'.join((user_name, pwd)),
            path=self.relative_uri,
        )

    @property
    def owner_basic_auth_url(self):
        return self.basic_auth_url(self.owner)

    def deploy_token_url(self, token):
        return self.basic_auth_url(token['username'], pwd=token['token'])

    @property
    def api_url(self):
        return '/'.join((
            self.heptapod.url,
            'api', 'v4', 'projects',
            '/'.join((self.group.full_path, self.name)).replace('/', '%2F')
        ))

    @property
    def fs_common_path(self):
        """Common abspath on Heptapod server FS (not ending with .hg nor .git)

        Meaningful only for those tests that require file system access.
        Relies on knowledge of internal GitLab details that may well change.
        (since these are tests, we would notice quickly).
        """
        disk_path = getattr(self, '_disk_path', None)
        if disk_path is not None:
            return disk_path

        if not self.is_legacy:
            sha2 = hashlib.sha256(b'%d' % self.id).hexdigest()
            rpath = '@hashed/%s/%s/%s' % (sha2[:2], sha2[2:4], sha2)
            disk_path = os.path.join(self.heptapod.repositories_root, rpath)
        else:
            disk_path = '/'.join((self.heptapod.repositories_root,
                                  self.group.full_path, self.name))
        self._disk_path = disk_path
        return disk_path

    @property
    def fs_path(self):
        """Path to the Mercurial repo on Heptapod server file system."""
        return self.fs_common_path + '.hg'

    @property
    def fs_path_git(self):
        """Path to the Git repo on Heptapod server file system."""
        return self.fs_common_path + '.git'

    def _change_storage(self, legacy):
        label = 'legacy' if legacy else 'hashed'
        if legacy is self.is_legacy:
            logger.warn("_change_storage: project %d is already on %s storage",
                        self.id, label)
        rake_cmd = 'rollback_to_legacy' if legacy else 'migrate_to_hashed'
        self.heptapod.rake('gitlab:storage:' + rake_cmd,
                           'ID_FROM=%d' % self.id,
                           'ID_TO=%d' % self.id)
        self.is_legacy = legacy
        self._disk_path = None

        # we're not inconsistent
        assert ('@hashed' in self.fs_path) is (not legacy)

        wait_assert(
            lambda: self.heptapod.execute(('test', '-e', self.fs_path))[0],
            lambda code: code == 0,
            timeout_factor=12,
            msg="Repository %r not found in %s storage" % (self.fs_path, label)
        )

    def make_storage_legacy(self):
        self._change_storage(True)

    def make_storage_hashed(self):
        self._change_storage(False)

    def rake_migrate_native(self):
        self.heptapod.rake(
            f'heptapod:experimental:hg_migrate_native'
            f'[{self.id},true,{self.owner_user.name}]'
        )

    def api_migrate_native(self, run_again=False, user=None, check=True):
        if user is None:
            user = self.owner_user

        resp = api_request('POST', self, user, subpath='hg_migrate_native',
                           data=dict(run_again=run_again))
        if not check:
            return resp

        assert resp.status_code < 400

    def wait_assert_api_migrate_native(self, **kwargs):
        """Migrate to native through API and wait_assert for completion.

        We wait first for the change of VCS type to native, then for
        unarchival, because waiting for the project archival would be at
        risk to miss the whole process if it turns out to be real fast.

        Chances are high that the project is unarchived very soon after
        the change of VCS type, but that does not matter.

        This process is unreliable if this is a rerun, as the VCS type and
        archival status are exactly as expected at the end from the onset.
        Hence it is possible, if the test is much faster than the server that
        the wait returns before the process even starts.

        Run-again tests expecting a visible change should wait_assert this
        precise change for better reliability.
        """
        self.api_migrate_native(**kwargs)
        wait_assert(lambda: self.api_get_field('vcs_type'),
                    lambda vcs_type: vcs_type == 'hg')
        wait_assert(lambda: self.api_get_field('archived'),
                    lambda archived: not archived)

    def commit_page_url(self, commit_id):
        """Commit URL for regular (webdriver) navigation."""
        return '/'.join((self.url, 'commit', commit_id))

    def archive_url(self, gitlab_branch, file_ext):
        base_file_name = '-'.join([self.name] + gitlab_branch.split('/'))
        return '/'.join((self.url,
                         '-', 'archive',
                         gitlab_branch,
                         base_file_name + '.' + file_ext
                         ))

    def get_archive(self, gitlab_branch, fmt='tar'):
        """Retrieve a repository archive by URL.

        :param fmt: the wished format. At this point, this is directly
            mapped as a file extension in the request, and only the `tar`
            value is tested.
        :returns: the archive content, as a files-like object
        """
        resp = requests.get(self.archive_url(gitlab_branch, fmt))
        assert resp.status_code == 200
        return BytesIO(resp.content)

    def webdriver_get_raw_blob(self, path, revision, webdriver=None):
        """Use a webdriver to download raw file content at a given revision.

        :param revision: anything that the Rails app understands, in
           particular GitLab branch name and commit hash.
        """
        if webdriver is None:
            webdriver = self.owner_webdriver

        webdriver.get('/'.join((self.url, '-', 'raw', revision, path)))
        return raw_page_content(webdriver)

    def api_branches(self):
        """Retrieve and pre-sort branches info through the REST API."""
        resp = self.api_get(subpath='repository/branches')
        assert resp.status_code == 200
        branches = resp.json()
        return dict((branch.pop('name'), branch) for branch in branches)

    def api_default_branch(self):
        branch = self.api_get_field('default_branch')
        assert branch is not None
        return branch

    def api_branch_titles(self):
        """Keep only commit titles from `api_branches()`

        With a test scenario in which those titles are characterizing the
        commit uniquely, this is what's very often needed for assertions.
        """
        return extract_gitlab_branch_titles(self.api_branches())

    def wait_assert_api_branches(self, condition,
                                 msg="The given condition on GitLab branches "
                                 "was still not fulfilled after retrying "
                                 "for {timeout} seconds",
                                 **kwargs):
        """Assert some condition to become True on GitLab branches.

        Since the update of pushed or pruned branches is asynchronous and
        becomes even more so as GitLab progresses, this provides the means
        to retry several calls to :meth:`api_branches`.

        :param condition: any callable returning boolean that can take a
                          single argument, the payload of :meth:`api_branches`
        :param kwargs: passed to the underlying :func:`wait_assert` call.
        :returns: branches after the wait
        :raises: AssertionError if the condition doesn't become True before
                 timeout
        """
        return wait_assert(lambda: self.api_branches(), condition,
                           msg=msg, **kwargs)

    def wait_assert_api_branch_titles(self, expected, **kw):
        self.wait_assert_api_branches(
            lambda branches: (extract_gitlab_branch_titles(branches)
                              == expected),
            **kw,
        )

    def api_tags(self):
        """Retrieve and pre-sort tags info through the REST API."""
        resp = self.api_get(subpath='repository/tags')
        assert resp.status_code == 200
        tags = resp.json()
        return dict((tag.pop('name'), tag) for tag in tags)

    def api_protected_branches(self):
        resp = self.api_get(subpath='protected_branches')
        assert resp.status_code == 200
        return {br['name']: br for br in resp.json()}

    def api_unprotect_branch(self, branch_spec, check=True):
        """Unprotect the given branh specification.

        :param check: if ``True``, assert success. Otherwise return response.
        """
        resp = api_request('DELETE', self, self.owner_user,
                           subpath=protected_branch_api_subpath(branch_spec))
        if not check:
            return resp

        assert resp.status_code < 400

    def api_protect_branch(self, branch_spec, check=True, **access_levels):
        """Create branch protection or update level.

        :param access_levels: a :class:`dict` with :class:`ProjectAccess`
           values
        """
        self.api_unprotect_branch(branch_spec)

        data = {k: level.value for k, level in access_levels.items()}
        data['name'] = branch_spec
        resp = api_request('POST', self, self.owner_user,
                           subpath=protected_branch_api_subpath(),
                           data=data)

        if not check:
            return resp

        assert resp.status_code < 400

    def api_commit(self, sha, check=True):
        """Retrieve a commit by its SHA.

        The SHA is the native one to GitLab, typically obtained through the
        API. For Mercurial SHAs, it's usually simpler to just perform a pull.
        """
        resp = self.api_get(subpath='repository/commits/' + sha)
        if not check:
            return resp

        assert resp.status_code == 200
        return resp.json()

    def api_file_create(self, path, check=True, **data):
        """data is transferred directly into the JSON expected by the API."""
        data['file_path'] = path
        resp = self.api_post(
            subpath='repository/files/' + path,
            data=data)
        if not check:
            return resp

        assert resp.status_code < 400
        return resp.json()

    def api_file_update(self, path, check=True, **data):
        """data is transferred directly into the JSON expected by the API."""
        data['file_path'] = path
        resp = self.api_put(
            subpath='repository/files/' + path,
            data=data)
        if not check:
            return resp

        assert resp.status_code < 400
        return resp.json()

    def api_wiki_page_create(self, user=None, check=True, **data):
        resp = self.user_api_request('POST', user=user,
                                     subpath='wikis', data=data)
        if not check:
            return resp
        assert resp.status_code == 201
        return resp.json()

    def api_wiki_page_get(self, slug, user=None, check=True):
        resp = self.user_api_request('GET', user=user,
                                     subpath='wikis/' + slug)
        if not check:
            return resp
        assert resp.status_code == 200
        return resp.json()

    def api_wiki_page_update(self, slug, user=None, check=True, **data):
        resp = self.user_api_request('PUT', user=user,
                                     subpath='wikis/' + slug,
                                     data=data)
        if not check:
            return resp
        assert resp.status_code == 200
        return resp.json()

    def api_wiki_pages_list(self, user=None, check=True):
        resp = self.user_api_request('GET', user=user, subpath='wikis')
        if not check:
            return resp
        assert resp.status_code == 200
        return resp.json()

    def hg_wiki_url(self, user_name=None):
        """Return an authenticated URL suitage for hg pull/push.

        :param user_name: any user name known to :attr:`heptapod`
        """
        if user_name is None:
            user_name = self.owner
        return self.basic_auth_url(user_name) + '.wiki'

    def api_file_get(self, path, ref, content_only=True):
        """Retrieve a repository file through API.

        :param content_only: if ``True``, the response status code is checked
           and the content is extracted and returned as
           bytes. Otherwise the full HTTP response is returned.
        """
        resp = self.api_get(subpath='repository/files/' + path,
                            params=dict(ref=ref))
        if not content_only:
            return resp

        assert resp.status_code == 200
        return b64decode(resp.json()['content'])

    def api_file_delete(self, path, check=True, **data):
        data['file_path'] = path
        resp = self.api_delete(subpath='repository/files/' + path,
                               data=data)
        if not check:
            return resp

        assert resp.status_code < 400

    def webdriver_update_merge_request_settings(self, merge_method):
        driver = self.owner_webdriver
        driver.get('{url}/{group_path}/{project_name}/edit'.format(
            group_path=self.group.full_path,
            url=self.heptapod.url,
            project_name=self.name,
        ))

        elem = driver.find_element(
            By.XPATH,
            '//section[contains(@class, "merge-requests-feature")]'
            '//button[contains(text(), "Expand")]'
        )
        elem.click()

        wait_could_click_element(
            driver,
            lambda d: d.find_element(By.ID,
                                     'project_merge_method_' + merge_method))

        submit = driver.find_element(
            By.XPATH,
            '//section[contains(@class, "merge-requests-feature")]'
            '//input[@type="submit"]')
        submit.click()

    def api_update_merge_request_settings(self, merge_method):
        resp = api_request('PUT', self, self.owner_user,
                           data=dict(merge_method=merge_method))
        assert resp.status_code < 400

    def api_fork(self, user, group=None):
        data = dict(id=self.id)
        if group is None:
            group = UserNameSpace(user_name=user.name, heptapod=self.heptapod)
        else:
            data['namespace_id'] = group.id

        resp = self.user_api_request('POST',
                                     user=user, subpath='fork', data=data)
        assert resp.status_code < 400

        fork_info = resp.json()
        owner_info = fork_info.get('owner')
        if owner_info is not None:
            assert owner_info['username'] == user.name
        fork = self.__class__(
            heptapod=self.heptapod,
            group=group, name=fork_info['path'],
            id=fork_info['id'],
            vcs_type=fork_info.get('vcs_type'),
            owner=user.name)
        return wait_assert(
            lambda: fork,
            lambda proj: proj.api_get_field('import_status') == 'finished'
        )

    def webdriver_fork(self, user):
        driver = user.webdriver
        driver.get(self.url)
        wait_could_click_element(
            driver,
            lambda d: d.find_element(By.CSS_SELECTOR, 'a.fork-btn'))
        select = driver.find_element(By.NAME, 'namespace')
        select.click()
        wait_could_click_element(
            driver,
            lambda d: d.find_element(
                By.XPATH,
                '//select[@name="namespace"]'
                '//option[contains(text(), "%s")]' % user.name))

        submit = driver.find_element(
            By.XPATH,
            '//button[@type="submit" '
            'and @data-qa-selector="fork_project_button"]')
        submit.click()

        group = UserNameSpace(user_name=user.name, heptapod=self.heptapod)

        return wait_assert(
            lambda: self.api_retrieve(heptapod=self.heptapod,
                                      user_name=user.name,
                                      group=group,
                                      name=self.name,
                                      check=False),
            lambda fork: fork is not None
        )

    def api_create_merge_request(self, source,
                                 target='branch/default',
                                 **kwargs):
        """As the project owner, create a merge request through API.

        :param source: name of the source branch for GitLab, e.g.,
                    'topic/default/foo'
        :param target: name of the target branch for GitLab
        :param target_project: target :class:`Project` instance, defaults to
           ``self``, which is already the source project.
        :returns: numeric iid of the newly created MR
        """
        warnings.warn("Project.api_create_merge_request() is deprecated, "
                      "use MergeRequest.api_create() instead",
                      DeprecationWarning, stacklevel=2)
        return MergeRequest.api_create(self, source,
                                       target_branch=target,
                                       **kwargs).iid

    def api_create_deploy_token(self, name, scopes=('read_repository',)):
        """Will be available with GitLab 12.9."""
        resp = self.api_post(subpath='deploy_tokens',
                             data=dict(name=name,
                                       scopes=scopes))
        assert resp.status_code == 201
        return resp.json()

    def api_delete_deploy_token(self, token):
        """Will be available with GitLab 12.9."""
        self.api_delete(subpath='deploy_tokens',
                        data=dict(token_id=token['id']))

    def webdriver_create_deploy_token(self, name):
        """Create a deploy token with Selenium

        The API doesn't exist before GitLab 12.9.
        This method does not work on GitLab 13.12 anymore. Probably not
        a big deal, but we can use the API instead for all current use cases.

        :param name: required by GitLab, is only a human intended description
        """
        driver = self.owner_webdriver
        driver.get(self.url + '/-/settings/repository')
        wait = webdriver_wait(driver)

        section_xpath = '//section[contains(@class, "deploy-tokens-settings")]'
        expand_button = driver.find_element(
            By.XPATH,
            section_xpath + '//button[contains(text(), "Expand")]')
        expand_button.click()
        wait.until(lambda d: d.find_element(
            By.XPATH,
            section_xpath + '//button[contains(text(), "Collapse")]'
        ).is_displayed)

        def name_elt(d):
            return d.find_element(By.ID, 'deploy_token_name')

        wait.until(lambda d: name_elt(d).is_displayed())
        name_elt(driver).send_keys(name)
        read_repo_scope = driver.find_element(
            By.ID,
            'deploy_token_read_repository')
        read_repo_scope.click()
        submit = driver.find_element(
            By.XPATH,
            '//form[@id="new_deploy_token"]'
            '//input[@type="submit"]')
        submit.click()

        def value_for_id(elt_id):
            return driver.find_element(By.ID, elt_id).get_attribute('value')

        return dict(
            username=value_for_id('deploy-token-user'),
            token=value_for_id('deploy-token'),
        )

    def grant_member_access(self, user_name, level):
        """Grant given user the given access level.

        It doesn't matter whether the user is already a member or not: this
        method abstracts over it.

        This method is idempotent.
        """
        assert level in ProjectAccess

        user_id = self.heptapod.get_user(user_name).id
        resp = self.api_get(subpath='members/%d' % user_id)
        if resp.status_code == 404:
            subpath = 'members'
            meth = 'POST'
        else:
            subpath = 'members/%d' % user_id
            meth = 'PUT'

        resp = api_request(meth, self, self.owner_user,
                           subpath=subpath,
                           data=dict(user_id=user_id, access_level=int(level)))
        assert resp.status_code < 400

    def load_tarball(self, tarball_path):
        """Replace server-side repository files by the contents of tarball.

        This should be used right after the project creation.

        :param Path tarball_path: path to an uncompressed tar
            archive containing `hg` and `git`. These
            will be renamed to the values of `self.fs_path`` and
            ``self.fs_path_git``.
        """
        if not self.heptapod.fs_access:
            raise NotImplementedError(
                "Can't use load_tarball() without filesystem access")

        # initialize repository
        # GitLab needs a first clone or push to consider the Git repository to
        # exist. Otherwise, local pushes such as what the heptapod sync hook
        # does just fail, with an error about the Git repo existence.
        tmpdir = py.path.local.mkdtemp()
        try:
            LocalRepo.clone(self.owner_basic_auth_url, tmpdir.join('clone'))
        finally:
            tmpdir.remove()

        heptapod = self.heptapod
        # using a temporary space in same mount point and unique enough
        srvtmpdir = self.fs_common_path + '.tmp'
        heptapod.run_shell(['rm', '-rf',
                            self.fs_path, self.fs_path_git, srvtmpdir])
        heptapod.run_shell(['mkdir', '-p', srvtmpdir])
        heptapod.put_archive(srvtmpdir, tarball_path)
        heptapod.run_shell(['mv', srvtmpdir + '/hg', self.fs_path])
        heptapod.run_shell(['mv', srvtmpdir + '/git', self.fs_path_git])

    def get_hgrc(self, managed=False):
        """Return repo's server-side HGRC, as lines, uid and gid

        :param managed: if ``True``, the contents returned are those of the
                        file managed by the Rails app, introduced
                        for heptapod#165
        """
        hgrc_path = '/'.join((self.fs_path, '.hg',
                              'hgrc.managed' if managed else 'hgrc'))
        return self.heptapod.get_file_lines(hgrc_path)

    def put_hgrc(self, lines, file_path='hgrc'):
        """Replace repo's server-side HGRC with given lines.

        The lines have to include LF, same as with `writelines()`.

        :param str file_path: relative path from the `.hg` directory of the
           repo for the file to rewrite. Allows to work on a different file
           than the main HGRC. Useful to manage secondary files (inclusions,
           clonebundles, configexpress…)
        """
        repo_inner_path = '/'.join((self.fs_path, '.hg'))

        return self.heptapod.put_file_lines(
            os.path.join(repo_inner_path, file_path), lines)

    def extend_hgrc(self, *lines):
        """Append given lines to repo's server-side HGRC

        The lines don't have to be newline-terminated.
        """
        hgrc_lines = self.get_hgrc()
        # just in case original hgrc isn't new-line terminated
        hgrc_lines.append('\n')
        hgrc_lines.extend(l + '\n' for l in lines)
        self.put_hgrc(hgrc_lines)

    def hg_config(self, section=None):
        """Return Mercurial configuration item, as really seen by hg process.

        In other words, this isn't inference on the contents of the various
        HGRC, it can be used to test propagation of config entries.

        :return: if ``section`` is passed, a simple ``dict``, otherwise a
                 ``dict`` of ``dicts``. End values are always strings.
        """
        cmd = [self.heptapod.hg_executable, '-R', self.fs_path,
               '--pager', 'false',
               'config']
        if section is not None:
            cmd.append(section)

        code, out = self.heptapod.execute(cmd, user='git')
        config = {}
        if out is None:
            return config
        for l in out.splitlines():
            print(l)
            fullkey, val = l.split('=', 1)
            section, key = fullkey.split('.', 1)
            if section is not None:
                config[key] = val
            else:
                config.setdefault(section, {})[key] = val
        return config

    def api_destroy(self, allow_missing=False, timeout_factor=1,
                    as_user=None):
        if as_user is None:
            as_user = self.owner  # just the owner name (str)

        if isinstance(as_user, str):
            as_user = self.heptapod.get_user(as_user)

        def delete(as_user):
            resp = requests.delete(self.api_url,
                                   headers=as_user.token_headers())
            if allow_missing and resp.status_code == 404:
                return
            print("DELETE request response: %r" % resp.json())
            assert resp.status_code == 202
            return resp

        # Even though the deletion meaning that the repos are just
        # mv'ed on the filesystem, it is still async
        return wait_assert(
            lambda: delete(as_user),
            lambda _resp: self.api_get().status_code == 404,
            msg="Project %r was not destroyed (still accessible) after "
            "{timeout} seconds",
        )

    def webdriver_destroy(self, skip_missing=False):
        driver = self.owner_webdriver
        driver.get(self.url + '/edit')
        if skip_missing and '404' in driver.title:
            return

        assert 'General' in driver.title
        assert 'Settings' in driver.title
        assert self.name in driver.title

        wait = webdriver_wait(driver)

        elem = driver.find_element(
            By.XPATH,
            '//section[contains(@class, "advanced-settings")]'
            '//button[contains(text(), "Expand")]'
        )
        elem.click()
        wait.until(lambda d: d.find_element(
            By.XPATH,
            '//section[contains(@class, "advanced-settings")]'
            '//button[contains(text(), "Collapse")]'
        ).is_displayed)

        wait.until(could_click_element(lambda d: d.find_element(
            By.XPATH,
            # a <button> inside a <form> that contains
            # an <input name="_method" value="delete">
            '//form[input[@name="_method" and @value="delete"]]/button'
        )))

        # a global find by name or id retrieves the wrong element (despite
        # id being supposed to be unique)
        modal = driver.find_element(
            By.XPATH,
            '//div[contains(@id, "delete-project-modal")]')
        confirm_input = modal.find_element(By.NAME, 'confirm_name_input')

        confirm_input.send_keys(modal.find_element(By.TAG_NAME, 'code').text)
        confirm_button = modal.find_element(
            By.CSS_SELECTOR,
            'button.js-modal-action-primary')
        confirm_button.click()

        deletion_banner_re = r"Project '\w+ / {}' will be deleted.".format(
            self.name)

        # storing page source in a variable to allow for investigation
        # in case the regexp doesn't match and we've not been redirected
        # yet. (we don't know at this point if the regexp is appropriate
        # for GitLab 13.3)
        html = driver.page_source
        if re.search(deletion_banner_re, html) is None:
            # It can happen that the deletion has already been started
            # and we have been redirected to the project list.
            assert 'dashboard/projects' in driver.current_url
            # We follow redirects a bit too fast for the deletion always
            # to be fully done. Watching the dashbord page while the repos
            # are already deleted on the filesystem, yet the project
            # still exists in database is an error cause.
            wait_assert(
                lambda: self.api_get(),
                lambda resp: resp.status_code == 404,
                msg="Project was still accessible in API after {timeout}",
            )
            driver.get(driver.current_url)
            assert_webdriver_not_error(driver)

    @classmethod
    def api_retrieve(cls, heptapod, user_name, group, name, check=True):
        """Retrieve Project instance if `id` is not known.

        :param user_name: name of an user having enough access to retrieve
                          project info and in particular owner name
        """
        project = cls(heptapod=heptapod,
                      owner=user_name,  # temporary
                      group=group,
                      name=name)
        resp = project.api_get()
        if check:
            assert resp.status_code == 200
        elif resp.status_code == 404:
            return

        data = cls.assert_project_info(resp, heptapod)
        project.vcs_type = data['vcs_type']
        project.id = data['id']
        creator_id = data.get('creator_id')
        owner = data.get('owner')
        if owner is not None:
            project.owner = data['owner']['username']
        elif creator_id is not None:
            project.creator_id = creator_id
        else:
            raise LookupError("No owner/creator information for %r", project)

        return project

    @classmethod
    def webdriver_create(cls, heptapod, user_name, project_name, **kwargs):
        """Create a new project with the webdriver for given user
        """
        group = UserNameSpace(heptapod=heptapod, user_name=user_name)
        driver = heptapod.get_user(user_name).webdriver
        driver.get('{url}/projects/new'.format(url=heptapod.url))
        cls.webdriver_new_project_submit(driver, project_name, **kwargs)
        project = cls.api_retrieve(heptapod, user_name, group, project_name)

        project.wait_hg_availability()
        return project

    @classmethod
    def webdriver_import_url(cls, heptapod, user, project_name, url,
                             wait_import_url_validation=True,
                             timeout_factor=3,
                             **kwargs):
        """Import a new project with the webdriver for given user,

        The project will be in the user personal namespace.
        """
        group = UserNameSpace(heptapod=heptapod, user_name=user.name)
        driver = user.webdriver
        driver.get(heptapod.url + '/projects/new#import_project')
        url_button = driver.find_element(By.CSS_SELECTOR,
                                         'button.js-import-git-toggle-button')
        url_button.click()

        def fill_url(driver):
            url_input_spec = (By.NAME, 'project[import_url]')
            elem = driver.find_element(*url_input_spec)
            elem.send_keys(url)
            if kwargs.get('vcs_type'):
                # must wait for VCS type selection to be taken into account
                # (same comment as below about proper waiting)
                time.sleep(BASE_TIMEOUT / 10)
            elem.send_keys(Keys.TAB)
            if wait_import_url_validation:
                # proper waiting is hard to implement, because if validation
                # succeeds, nothing changes on the page, so there's no element
                # to wait on (typically in redisplay it would be a wait for
                # staleness of old element, then wait for apparition of new
                # element. There are several recipes online, most of them
                # relying on `jQuery.active`, which we can't assume to work.
                # Some other methods inject JavaScript that seems to add
                # further callbacks to XHRs… Too complicated for now.
                # So this is lame, but good enough to reproduce heptapod#524
                time.sleep(BASE_TIMEOUT / 10)

        cls.webdriver_new_project_submit(driver, project_name,
                                         additional_driver_actions=fill_url,
                                         select_blank_project=False,
                                         # the hidden init README checkbox
                                         # is selected (and ignored), don't
                                         # try to toggle it.
                                         init_with_readme=True,
                                         assert_created_user_feedback=False,
                                         **kwargs)
        project = cls.api_retrieve(heptapod, user.name, group, project_name)
        project.wait_assert_import(timeout_factor=timeout_factor)
        return project

    def wait_hg_availability(self):
        wait_assert(lambda: self.owner_get(params=dict(cmd='capabilities')),
                    lambda resp: resp.status_code == 200)

    @classmethod
    def webdriver_new_project_submit(cls, driver, project_name,
                                     select_blank_project=True,
                                     additional_driver_actions=None,
                                     assert_created_user_feedback=True,
                                     vcs_type=None,
                                     import_url_validation=False,
                                     init_with_readme=False):
        """Create a new project while already on the "New project" page.

        This will put the project in the default namespace proposed by
        the "New project" page. We could add a `group` kwarg later on.
        """
        assert 'GitLab' in driver.title
        assert 'New Project' in driver.title

        if select_blank_project:
            webdriver_wait(driver).until(
                could_click_element(
                    lambda d: d.find_element(
                        By.XPATH,
                        '//a[@href = "#blank_project"]')))

        # the project path (slug) attribute is derived automatically from
        # its (human readable) name. The names given in these tests all
        # transform identically into paths.
        # In GitLab 12.10 the converse is true: if only the path is given,
        # the name is derived from it with some prettyfication (capitalization,
        # spaces). Therefore it is better for our needs to provide the name
        # only.
        elem = driver.find_element(By.NAME, 'project[name]')
        elem.send_keys(project_name)
        # If we wanted later on to make sure the path is the intended
        # one, we'd have to clear the `project[path]` field first, instead
        # of *adding* path after what was just prefilled due to
        # the setting of `project[name]`

        # in GitLab 14.2 the initialization with README is selected by default
        init_readme_checkbox = driver.find_element(
            By.ID,
            'project_initialize_with_readme')
        selected = init_readme_checkbox.get_attribute('selected') == 'true'
        if not(init_with_readme) is selected:
            init_readme_checkbox.click()

        if vcs_type is not None:
            option = driver.find_element(By.XPATH,
                                         '//select[@id="project_vcs_type"]'
                                         '//option[@value="%s"]' % vcs_type)
            option.click()

        if additional_driver_actions is not None:
            additional_driver_actions(driver)

        elem = driver.find_element(
            By.XPATH,
            '//input[@type="submit" '
            'and @value="Create project"]')
        elem.click()
        webdriver_wait(driver).until(
            lambda d: 'will automatically refresh' not in d.page_source)

        if not assert_created_user_feedback:
            return

        wait_assert_in_page_source(
            driver, f"Project '{project_name}' was successfully created.")

    def webdriver_retrieve(cls, driver, heptapod, group, name,
                           wait_availability=False):
        """Use a driver to retrieve the project if it exists.

        If the project is not found (could happen with authnz problems),
        then `None` is returned.
        """
        project = Project(heptapod=heptapod, group=group, name=name)
        resp = project.session_api_get(driver)

        if resp.status_code == 404:
            return

        data = resp.json()
        project.owner = data['owner']['username']
        project.id = data['id']
        return project

    @classmethod
    def api_create(cls, heptapod, user_name, project_name,
                   group=None, timeout_factor=3, **data):
        """Create project using the RESTful API.

        :param user_name: name of the user to run the import with
        :param group: name of the project group (by default, will be
                      user personal space.
        :param data: for extra data in the POST API request that performs
                     the creation
        :returns: Project instance
        """
        if group is None:
            group = UserNameSpace(heptapod=heptapod, user_name=user_name)
        else:
            data['namespace_id'] = group.id
        group_path = group.full_path

        headers = heptapod.get_user(user_name).token_headers()

        data['name'] = project_name
        url = heptapod.url + '/api/v4/projects'

        data.setdefault('vcs_type', 'hg' if heptapod.hg_native else 'hg_git')

        def post():
            return requests.post(url, headers=headers, data=data)

        def needs_retry(response):
            if response is None:
                return True
            if response.status_code != 400:
                return False
            structured = response.json()
            message = structured.get('message', {})

            if ['has already been taken'] in (message.get('name'),
                                              message.get('path')):
                # clean up if we can
                try:
                    route = '/'.join((group_path, project_name))
                    logger.warn("GitLab Route %r has already been taken, "
                                "attempting to remove it", route)
                    heptapod.force_remove_route(route)
                except NotImplementedError:
                    pass

                # trigger retry even if we can't clean up from here:
                # on GitLab 12, there's a short timespan after it stops
                # telling us to retry but before it has fully cleaned up routes
                return True

            if any('try again' in l for l in message.get('base', ())):
                logger.info(
                    "Got explicit retry response: %r with headers %r",
                    structured, response.headers)
                return True

        resp = wait_assert(post, lambda resp: not needs_retry(resp),
                           timeout_factor=timeout_factor,
                           msg="Project could not be created "
                           "in {timeout} seconds",
                           )

        logger.debug("Creation request response: %r", resp.json())
        assert resp.status_code in (200, 201)

        vcs_type = data['vcs_type']
        proj_info = cls.assert_project_info(resp, heptapod,
                                            owner_name=user_name,
                                            vcs_type=vcs_type)

        return cls(heptapod=heptapod,
                   group=group, name=project_name,
                   id=proj_info['id'],
                   vcs_type=vcs_type,
                   hg_git_repo_expected=not heptapod.hg_without_git,
                   owner=user_name)

    @classmethod
    def assert_project_info(cls, api_resp, heptapod,
                            owner_name=None,
                            vcs_type=None):
        """Check that defining fields are as expected in an API response

        :returns: project info dict, as given by API response
        """
        proj_info = api_resp.json()
        if owner_name is not None:
            owner_info = proj_info.get('owner')
            if owner_info is not None:
                assert owner_info['username'] == owner_name

        if vcs_type is not None:
            assert proj_info['vcs_type'] == vcs_type
        if proj_info['vcs_type'] == 'hg':
            # either the `hgitaly1` value has been finally removed from the
            # API response or it is False, as expected since we don't
            # test the HGitaly1 mode any more.
            assert not proj_info.get('hgitaly1')
            assert proj_info['hg_without_git'] == heptapod.hg_without_git

        return proj_info

    @classmethod
    def api_import_url(cls, heptapod, user, project_name, url,
                       group=None, **wait_assert_kw):
        """Import project from URL, wait for completion and returns it.

        :param user_name: name of the user to run the import with
        :param group: name of the project group (by default, will be
                      user personal space.
        :returns: import_status ('failed' or 'finished'), import_error,
                  Project instance
        """
        project = cls.api_create(heptapod, user.name, project_name,
                                 group=group,
                                 import_url=url)
        project.wait_assert_import(**wait_assert_kw)
        return project

    def wait_assert_import(self, timeout_factor=3):
        unexpected_statuses = (None, 'none')
        terminal_statuses = ('finished', 'failed')

        def status_is_terminal(proj_info):
            status = proj_info['import_status']
            if status in terminal_statuses:
                return True
            assert status not in unexpected_statuses

        proj_info = wait_assert(
            lambda: self.api_get_info(),
            lambda proj_info: proj_info['import_status'] in terminal_statuses,
            msg="Import was still running after {timeout} seconds",
            timeout_factor=timeout_factor,
        )
        if proj_info['import_status'] == 'failed':
            raise AssertionError(proj_info.get('import_error'))

        print("Project information after import or timeout: \n%r" % proj_info)

    @classmethod
    def api_import_tarball(cls, heptapod, user_name,
                           project_name, tarball_fobj,
                           group=None,
                           timeout_factor=3,
                           retry_wait_factor=0.05,
                           ):
        """Import project from a tarball, wait for completion and returns it.

        :param user_name: name of the user to run the import with
        :param group: name of the project group (by default, will be
                      user personal space.
        :param tarball_fobj: open file-like object for the tarball to import
        :returns: Project instance
        """
        user = heptapod.get_user(user_name)
        headers = user.token_headers()
        data = dict(path=project_name)

        if group is None:
            group = UserNameSpace(heptapod=heptapod, user_name=user_name)
        else:
            data['namespace_id'] = group.id

        resp = requests.post(heptapod.url + '/api/v4/projects/import',
                             headers=headers,
                             data=data,
                             files=dict(file=('proj.tar.gz', tarball_fobj)))
        assert resp.status_code < 400

        proj_info = resp.json()
        proj = cls(heptapod=heptapod,
                   group=group, name=project_name,
                   id=proj_info['id'],
                   # proj_info['vcs_type'] is at this point just a default one
                   # that the import process will correct from information
                   # in the tarball.
                   owner=user_name)

        resp = wait_assert(
            lambda: proj.api_get(subpath='import'),
            lambda resp: (resp.status_code < 400
                          and resp.json()['import_status'] == 'finished'),
            timeout_factor=timeout_factor,
            retry_wait_factor=retry_wait_factor,
            msg="Import not finished in {timeout} seconds",
        )
        proj.vcs_type = resp.json()['vcs_type']
        return proj

    @classmethod
    def api_list(cls, heptapod, search_as='root', **params):
        """List the projects owned by the given user

        Does not support batching at this point, hence will work only
        if the result set fits in the first page.

        Various filtering criteria can be applied, hence it is more a search
        than a listing (notably ``search=something`` is possible)

        :param search_as: login name of the user to perform the listing.
                          Must be one of those registered to the
                          :class:`Heptapod` instance.
        """
        resp = requests.get(
            heptapod.api_url + '/projects',
            headers=heptapod.get_user(search_as).token_headers(),
            params=params
        )
        return resp

    def api_transfer(self, group):
        # API for regular users (on the project) has been introduced in
        # GitLab 11.1 (see https://docs.gitlab.com/ce/api/projects.html)
        # but in current GitLab, root can do it
        heptapod = self.heptapod
        resp = requests.post(
            '{api}/groups/{group.id}/projects/{proj.id}'.format(
                api=heptapod.api_url, group=group, proj=self),
            headers=heptapod.get_user('root').token_headers(),
        )
        if resp.status_code < 400:
            # let's have in particular correct subsequent URLs
            self.group = group
            if self.is_legacy:
                # path on disk actually changes for legacy projects
                self._disk_path = None
        return resp

    def api_export(self, target_fobj,
                   timeout_factor=3, retry_wait_factor=0.05
                   ):
        """Export the project in a given file

        This is synchronous from caller's perspective.

        :target_fobj: a writable, binary, file-like object.
        """
        resp = self.api_post(subpath='export')
        assert resp.status_code == 202

        wait_assert(
            lambda: self.api_get(subpath='export'),
            lambda resp: (resp.status_code < 400
                          and resp.json()['export_status'] == 'finished'),
            timeout_factor=timeout_factor,
            retry_wait_factor=retry_wait_factor,
            msg="Export not finished in {timeout} seconds",
        )

        resp = self.api_get(subpath='export/download', stream=True)
        assert resp.status_code == 200
        for chunk in resp.iter_content(chunk_size=8192):
            target_fobj.write(chunk)

    def only_specific_runners(self):
        resp = self.api_edit(shared_runners_enabled=False)
        assert resp.status_code < 400

        if resp.json()['namespace']['kind'] != 'group':
            return

        driver = self.owner_webdriver

        driver.get(self.url + '/-/settings/ci_cd')
        elem = driver.find_element(
            By.XPATH,
            '//section[@data-qa-selector="runners_settings_content"]'
            '//button[contains(text(), "Expand")]'
        )
        elem.click()

        webdriver_wait(driver).until(
            could_click_element(
                lambda d: d.find_element(
                    By.XPATH,
                    '//a[contains(@href, "/toggle_group_runners")]')))

    def get_job_artifacts(self, job):
        """Return filename, artifacts (archive) as bytes."""
        resp = self.api_get(subpath='jobs/%d/artifacts' % job['id'])
        assert resp.status_code == 200
        # Finding file name depends on redirections (S3 etc.)
        # if we need something more precise, the 302 response is available
        # in resp.history
        if resp.url.startswith(self.heptapod.url):
            # no redirection, filename is in `Content-Disposition`
            # Not fully compliant filename parsing.
            # Full specification: RFC 6266.
            # there's an unsupported lib on PyPI for this (rfc6266)
            m = re.match(r'attachment; filename="(.*?)";',
                         resp.headers['Content-Disposition'])
            assert m is not None
            filename = m.group(1)
        else:
            # redirection, e.g, to S3 storage backend.
            # If needed, we could check it is really S3 by looking
            # for `AWSAccessKeyId` in query params.
            filename = urlparse(resp.url).path.rsplit('/', 1)[-1]

        return filename, resp.content

    def api_jobs(self, check=True):
        """Return CI jobs, per commit ID.

        The result is a :class:`dict` whose keys are GitLab commit IDs
        (Git or Mercurial SHA depending on VCS type) and values lists
        of jobs, expressed themselves as data dicts.
        """
        resp = self.api_get(subpath='jobs')
        if not check:
            return resp

        assert resp.status_code == 200
        jobs = {}
        for job in resp.json():
            jobs.setdefault(job['commit']['id'], []).append(job)
        return jobs

    def api_create_mirror(self, **params):
        return ProjectMirror.api_create(self, **params)

    def api_list_mirrors(self, check=True):
        """Return a list of dicts, not of ProjectMirror instances.

        The reason is that ProjectMirror currently does not keep more
        state than that's needed for identification.
        """
        resp = self.api_get(subpath='remote_mirrors')
        if not check:
            return resp

        assert resp.status_code < 300
        return resp.json()

    def assert_no_unwanted_git_repo(self):
        if self.vcs_type != 'hg' or self.hg_git_repo_expected:
            return

        heptapod = self.heptapod
        if not heptapod.fs_access:
            # we can't assert
            return

        assert not heptapod.path_exists(self.fs_path_git)

    def __enter__(self):
        return self

    def __exit__(self, *exc_args):
        # let's give a bit more time than usual: this is used for
        # cleanups and the instance could be recovering from various errors
        self.api_destroy(allow_missing=True, timeout_factor=3)
        return False



# Copyright 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import attr
import logging
import requests
import time
import threading

from gitlab import Gitlab  # python-gitlab
from selenium.webdriver.common.by import By


from .namespace import UserNameSpace
from .selenium import (
    webdriver_wait_get,
)
from .session import (
    login_as_root,
    sign_in_page_login,
)

logger = logging.getLogger(__name__)


def default_password(username):
    return username + '.' * (8 - len(username))


@attr.s
class User(object):
    name = attr.ib()
    id = attr.ib()
    heptapod = attr.ib(repr=False)
    email = attr.ib(default=None, repr=False)
    password = attr.ib(default=None, repr=False)
    token = attr.ib(default=None, repr=False)

    vcs_token_only = attr.ib(default=False, repr=False)
    """Are VCS HTTP operations to be made only with the access token?

    If two-factor authentication is activated, or if the user can only log
    through SSO, only private tokens can enable VCS HTTP push and pull.
    """

    thread_locals = attr.ib(factory=threading.local, repr=False)
    webdrivers_all_threads = attr.ib(factory=list, repr=False)
    _python_gitlab_client = attr.ib(default=None, repr=False)

    @classmethod
    def create(cls, heptapod, name,
               password=None,
               email=None,
               fullname=None, **kw):
        if password is None:
            # requirements on password length are also enforced in the API
            password = default_password(name)
        if fullname is None:
            fullname = name
        if email is None:
            email = name + '@heptapod.example.net',

        resp = requests.post(
            '/'.join((heptapod.api_url, 'users')),
            data=dict(name=fullname,
                      username=name,
                      password=password,
                      reset_password=False,
                      email=email,
                      skip_confirmation=True,
                      **kw
                      ),
            headers=heptapod.root_token_headers,
        )
        assert resp.status_code < 300, "User %r creation failed" % name
        user = cls(name=name, heptapod=heptapod, password=password,
                   email=email,
                   id=resp.json()['id'])
        user.store_in_heptapod()
        return user

    @classmethod
    def init_root(cls, heptapod, password):
        """Singled out because we have to use hardcoded values for root."""
        user = cls(id=1, heptapod=heptapod, name='root', password=password)
        user.store_in_heptapod()
        return user

    @classmethod
    def ensure(cls, heptapod, name, password=None, fullname=None):
        """Retrieve user or create it.

        :returns: `User` instance
        """
        if password is None:
            password = default_password(password)

        user = heptapod.users.get(name)
        if user is None:
            user = cls.search(heptapod, name)
            if user is None:
                user = cls.create(heptapod, name,
                                  password=password, fullname=fullname)
            user.store_in_heptapod()

        user.password = password
        user.ensure_private_token()
        return user

    @classmethod
    def search(cls, heptapod, name):
        """Lookup user in API and returns instance.

        The returned User instance doesn't know its password. Either it
        has an expected value, or it won't be suitable for webdriver
        authentication (API auth is usually done with access tokens).
        """
        resp = requests.get(
            '/'.join((heptapod.api_url, 'users')),
            params=dict(username=name),
            headers=heptapod.root_token_headers,
        )
        found = resp.json()
        if resp.status_code == 404 or not found:
            return

        assert len(found) == 1
        details = found[0]
        return cls(name=name,
                   id=details['id'],
                   email=details['email'],
                   heptapod=heptapod)

    def store_in_heptapod(self):
        self.heptapod.users[self.name] = self

    @property
    def personal_namespace(self):
        """Return a UserNameSpace instance for self."""
        return UserNameSpace(user_name=self.name, heptapod=self.heptapod)

    @property
    def webdriver(self):
        """Selenium webdriver, authentified as self.
        """
        driver = getattr(self.thread_locals, 'webdriver', None)
        if driver is not None:
            return driver

        thread_name = threading.current_thread().name

        logger.info(
            "Initializing a signed-in webdriver for user %r in thread %r",
            self.name, thread_name)

        # guaranteeing driver to be available for teardown as soon as created
        driver = self.thread_locals.webdriver = self.heptapod.new_webdriver()

        heptapod = self.heptapod
        if self.name == 'root':
            login_as_root(driver, heptapod, self.password)
        else:
            start = time.time()
            webdriver_wait_get(heptapod, driver, relative_uri='/users/sign_in')
            sign_in_page_login(driver, heptapod, self.name,
                               password=self.password)
            logger.info("Signed in user %s in %.2f seconds",
                        self.name, time.time() - start)
        # thank you, GIL
        self.webdrivers_all_threads.append((thread_name, driver))

        return driver

    def close_webdrivers(self):
        """Close current web driver for all threads."""
        for thread_name, driver in self.webdrivers_all_threads:
            logger.info("Closing webdriver of user %r for thread %r",
                        self.name, thread_name)
            driver.close()

        self.webdriver_all_threads = []
        self.thread_locals.webdriver = None

    def root_api_request(self, method, subpath=None, **data):
        """Perform a request on user API as root"""
        hepta = self.heptapod
        segments = [hepta.api_url, 'users', str(self.id)]
        if subpath is not None:
            segments.append(subpath)

        return requests.request(method, '/'.join(segments),
                                headers=hepta.root_token_headers,
                                data=data)

    def root_api_post(self, subpath=None, **data):
        return self.root_api_request('POST', subpath=subpath, **data)

    def delete(self, hard=True):
        self.close_webdrivers()
        hepta = self.heptapod

        unregistered = hepta.users.pop(self.name, None)
        if unregistered is not self:
            logger.warning("While unregistering user %r from Heptapod "
                           "instance, found %r registered in its place "
                           "instead", self.name, unregistered)

        resp = requests.delete(
            '/'.join((hepta.api_url, 'users', str(self.id))),
            data=dict(id=self.id, hard_delete=hard),
            headers=hepta.root_token_headers,
        )

        assert resp.status_code == 204, "Failed to delete %r" % self

    def edit(self, **data):
        resp = self.root_api_request('PUT', **data)
        assert resp.status_code == 200, "Failed to edit %r" % self

    def block(self):
        resp = self.root_api_post(subpath='block')
        assert resp.status_code == 201, "Failed to block %r" % self

    def unblock(self):
        resp = self.root_api_post(subpath='unblock')
        assert resp.status_code == 201, "Failed to unblock %r" % self

    def deactivate(self):
        resp = self.root_api_post(subpath='deactivate')
        assert resp.status_code == 201, "Failed to deactivate %r" % self

    def activate(self):
        resp = self.root_api_post(subpath='activate')
        assert resp.status_code == 201, "Failed to activate %r" % self

    def ensure_ssh_pub_key(self, pubkey, title='heptapod-tests'):
        """Upload the SSH public key if needed.

        Does not require any special privilege if a token for the user
        is already known
        """
        hepta = self.heptapod
        if self.token:
            keys_api_url = '/'.join((
                hepta.api_url, 'user', 'keys'))
            headers = self.token_headers()
        else:
            keys_api_url = '/'.join((
                hepta.api_url, 'users', str(self.id), 'keys'))
            headers = hepta.root_token_headers

        ls_resp = requests.get(keys_api_url, headers=headers)
        assert ls_resp.status_code == 200

        # check existence based on title. This is quite lame (we should
        # check fingerprint), but that's good enough for these first tests.
        for key_info in ls_resp.json():
            if key_info['title'] == title:
                return

        create_resp = requests.post(keys_api_url,
                                    data=dict(id=self.id,
                                              title=title,
                                              key=pubkey),
                                    headers=headers,
                                    )
        assert create_resp.status_code == 201
        assert create_resp.json()['title'] == title

    def delete_ssh_keys(self):
        """Delete all SSH keys of current user.

        Does not require any special privilege.
        """
        hepta = self.heptapod

        if self.token:
            keys_api_url = '/'.join((
                hepta.api_url, 'user', 'keys'))
            headers = self.token_headers()
        else:
            keys_api_url = '/'.join((
                hepta.api_url, 'users', str(self.id), 'keys'))
            headers = hepta.root_token_headers

        ls_resp = requests.get(keys_api_url, headers=headers)
        assert ls_resp.status_code == 200
        # check existence based on title. This is quite lame (we should
        # check fingerprint), but that's good enough for these first tests.
        for key_info in ls_resp.json():
            resp = requests.delete('%s/%d' % (keys_api_url, key_info['id']),
                                   headers=headers)
            assert resp.status_code == 204

    @property
    def ssh_command(self):
        """Return a SSH command to perform operations as the given user.
        """
        # IdentitiesOnly makes sure no other SSH key than the one
        # specified with -i are going to be used. Otherwise, current user
        # SSH keys could be attempted before the correct one, leading
        # to either too much auth failures, or use of the wrong key if
        # one happens to be also known by Heptapod
        return ' '.join(('ssh',
                         '-o', 'IdentitiesOnly=yes',
                         '-o', 'NoHostAuthenticationForLocalhost=yes ',
                         '-o', 'StrictHostKeyChecking=no',
                         '-i', self.ssh['priv']))

    def token_headers(self):
        """Return a *new* headers dict suitable for authenticated API acces.
        """
        return {'Private-Token': self.token}

    @property
    def basic_auth_url(self):
        url = self.heptapod.parsed_url
        return "{scheme}://{auth}@{netloc}".format(
            scheme=url.scheme,
            netloc=url.netloc,
            auth=':'.join((self.name, self.password)),
        )

    def api_get(self):
        """Perform GET on users API for self, using :attr:`token`.

        :return: response or `None` if :attr:`token` is None
        """
        token = self.token
        if token is None:
            return None

        return requests.get(self.heptapod.api_url + '/users/%d' % self.id,
                            headers=self.token_headers())

    def api_owned_groups(self):
        """List the groups owned by this user.

        TODO it lists actually the first page, which is expected to
        be good enough for the foreseeable future.
        """
        resp = requests.get(self.heptapod.api_url + '/groups',
                            params=dict(owned=True),
                            headers=self.token_headers(),
                            )
        assert resp.status_code < 400
        return resp.json()

    def ensure_private_token(self):
        """Generate a private token for already logged in user if needed.

        The generation involves obtaining a signed-in webdriver for the user.
        It is therefore costly.
        """
        resp = self.api_get()
        if resp is not None and resp.status_code == 200:
            logger.info("%r: reusing existing access token.", self)
            return

        heptapod = self.heptapod
        start = time.time()
        webdriver = self.webdriver
        webdriver.get(heptapod.url + '/-/profile/personal_access_tokens')
        name_elt = webdriver.find_element(
            By.XPATH,
            '//input[@name="personal_access_token[name]"]')
        name_elt.send_keys("heptapod-tests")

        for scope in webdriver.find_elements(
                By.XPATH,
                '//input[@type="checkbox"]'):
            # forcing click via JavaScript because on GitLab 14.10,
            # selenium complains with the "click intercepted" exception
            # and there is nothing obvious why (after inspections, screenshots,
            # etc).
            webdriver.execute_script("arguments[0].click();", scope)
        submit = webdriver.find_element(By.XPATH, '//input[@type="submit"]')
        submit.click()

        elem = webdriver.find_element(By.ID, 'created-personal-access-token')
        token = elem.get_attribute('value')
        self.token = token

        logger.info("%r: access token generated in %.2f seconds",
                    self, time.time() - start)

    @property
    def python_gitlab_client(self):
        """Return the API client from the python-gitlab library

        See https://python-gitlab.readthedocs.io
        """
        cl = self._python_gitlab_client
        if cl is None:
            cl = self._python_gitlab_client = Gitlab(self.heptapod.url,
                                                     private_token=self.token)
        return cl

    def __enter__(self):
        return self

    def __exit__(self, *exc_args):
        self.delete()
        return False



# Copyrightqhg 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import re
import time
from pytest import register_assert_rewrite

# Ask pytest to rewrite assertions in non-test modules
register_assert_rewrite("tests.utils.project", "tests.utils.session")


def unique_name(name):
    return '%s_%s' % (name, str(time.time()).replace('.', '_'))


def assert_message(message, expectations):
    """assert that a message matches expectations

    :param expectattions: any iterable whose elements can be either `bytes`,
       `str` or a compiled regexp.
       In the two former cases a simple `in` operator is used. In the latter
       case, the :meth:`search` is used.
       For caller convenience, a single value can also be used.

    Note: if expected is an empty string, the assertion is a no-op.
    """
    if isinstance(expectations, (str, bytes, re.Pattern)):
        expectations = (expectations, )

    for exp in expectations:
        if isinstance(exp, re.Pattern):
            if exp.search(message) is None:
                return False
        elif exp not in message:
            return False
    return True



import attr
import requests

from .wait import (
    wait_assert,
)

ARTIFACT_FORMATS = dict(archive='zip',
                        trace='raw',
                        metadata='gzip',
                        )


@attr.s
class Runner:
    heptapod = attr.ib()
    running_for = attr.ib()  # Group, Project or Heptapod
    description = attr.ib()
    id = attr.ib()
    token = None  # not displayed in repr etc

    @classmethod
    def api_register(cls, running_for, description):
        """Register the runner for a given scope of projects

        :param running_for: currently only a :class:`Project` instance, should
             eventually encompass :class:`Group` instances and the whole
             Heptapod.
        """
        heptapod = running_for.heptapod

        registration_token = running_for.api_get_field('runners_token')
        assert registration_token is not None

        resp = requests.post(heptapod.api_url + '/runners',
                             data=dict(token=registration_token,
                                       description=description,
                                       active=True,
                                       run_untagged=True,
                                       tag_list=[],
                                       ))
        assert resp.status_code == 201
        as_json = resp.json()

        runner = cls(heptapod=heptapod,
                     id=as_json['id'],
                     running_for=running_for,
                     description=description)
        runner.token = as_json['token']
        return runner

    def request_job(self):
        resp = requests.post(
            self.heptapod.api_url + '/jobs/request',
            json=dict(token=self.token,
                      info=dict(
                          executor='docker',
                          features=dict(
                              # always true (shells/abstract.go)
                              upload_multiple_artifacts=True,
                              upload_raw_artifacts=True,
                              refspecs=True,
                              artifacts=True,
                              artifacts_exclude=True,
                              multi_build_steps=True,
                              return_exit_code=True,
                              raw_variables=True,
                              cache=True,
                              masking=True,
                              # true for Docker executor
                              variables=True,
                              image=True,
                              services=True,
                              session=True,
                              terminal=True,
                              # shared seems to be set to True
                              # for shell, ssh and custom only.
                              shared=False,
                              # proxy seems to be set to True
                              # for kubernetes only.
                              # TODO evaluate if we shouldn't set
                              # it for PAAS Runner
                              proxy=False,
                          )),
                      ))
        assert resp.status_code < 400
        if resp.status_code != 204:  # or == 200
            return resp.json()

    def wait_assert_one_job(self, timeout_factor=3, retry_wait_factor=0.2,
                            msg="job not given after {timeout} seconds"):
        return wait_assert(lambda: self.request_job(),
                           lambda job: job is not None,
                           timeout_factor=timeout_factor,
                           retry_wait_factor=retry_wait_factor,
                           msg=msg,
                           )

    def wait_assert_jobs(self, number, **kwargs):
        """Fetches the given number of jobs, asserting their presence.

        GitLab workers will produce the jobs asychronously (that's
        why we wait). Hence the ordering of the results makes no sense.

        :param kwargs: passed down to each calls to `wait_assert()`.
        """
        return [
            self.wait_assert_one_job(
                msg="Got only %d job(s) instead of the expected %d "
                "after waiting for {timeout} seconds per job" % (i, number),
                **kwargs)
            for i in range(number)]

    def api_get(self):
        return requests.get(
            self.heptapod.api_url + '/runners/' + str(self.id),
            # Runner token can't be used for this
            headers={'Private-Token': self.running_for.owner_token})

    def api_delete(self):
        resp = requests.delete(self.heptapod.api_url + '/runners',
                               data=dict(token=self.token))
        assert resp.status_code == 204

    def upload_artifact(self, job, fobj, artifact_type='archive'):
        artifact_format = ARTIFACT_FORMATS[artifact_type]
        resp = requests.post(
            '{base_url}/jobs/{id}/artifacts'.format(
                base_url=self.heptapod.api_url,
                **job),
            params=dict(artifact_format=artifact_format,
                        artifact_type=artifact_type),
            headers={'JOB-TOKEN': job['token']},
            files=dict(file=fobj))
        assert resp.status_code < 400
        return resp.json()

    def __enter__(self):
        return self

    def __exit__(self, *exc_args):
        self.api_delete()
        return False


def job_variables(acquired_job):
    """Return job variables as a :class:`dict`.

    :param acquired_job: same :class:`dict` as acquired by a Runner
    """
    return {v['key']: v['value'] for v in acquired_job['variables']}



# Copyright 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os
import time

logger = logging.getLogger(__name__)

BASE_TIMEOUT = int(os.environ.get('HEPTAPOD_TESTS_BASE_TIMEOUT', '10').strip())


def log_base_timeout():
    logger.info("Base timeout is set to %r seconds", BASE_TIMEOUT)


def wait_assert(get_info, condition,
                before_retry=None,
                timeout_factor=1,
                retry_wait_factor=0.02,
                msg="Condition not fulfilled after {timeout} seconds"):
    """Assert some condition to become True after some tries.

    Many things are asynchronous in GitLab and moreso with time.

    :param get_info: callable with no arguments
    :param condition: callable that will passed the result of `get_info`
                      as its unique arguments
    :param before_retry: callable that will be invoked on retries only, with
      the result of `get_info`. It can be used to log the retry, or to
      recreate the proper conditions to call `get_info` again (if there were
      side effects).
    :param timeout_factor: how many times :const:`BASE_TIMEOUT` to wait until
       the condition is fulfilled.
    :param retry_wait_factor: time to wait between retries, expressed in
       multiples of the computed timeout.
    :param msg: format string used when raising AssertionError
    :returns: the latest list returned by `get_info`, guaranteed to fulfill
              `condition`
    :raises: AssertionError with the formatted given `msg`.
    """
    fulfilled = False
    timeout = timeout_factor * BASE_TIMEOUT
    retry_wait = timeout * retry_wait_factor
    start = time.time()
    while not fulfilled and time.time() - start < timeout:
        info = get_info()
        fulfilled = condition(info)
        if not fulfilled:
            if before_retry is not None:
                before_retry(info)
            time.sleep(retry_wait)

    if not fulfilled:
        raise AssertionError(msg.format(timeout=timeout))
    return info



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import requests


def api_request(method, obj, user, subpath='', **kwargs):
    """Perform a simple API HTTP request as the given user on given obj.


    :param method: the HTTP method to use, same as in `requests.request`.
    :param obj: the GitLab/Heptapod object, e.g. a `Project` instance.
       It must implement have an `api_url` attribute or property.
    :param user: a class:`User` instance.

    The full URL is made of the API URL of the object, together with
    the given subpath (example 'merge_requests/1').

    Appropriate authentication headers are added on the fly.

    All kwargs are passed to `requests.request()`
    """
    headers = kwargs.pop('headers', {})
    headers['Private-Token'] = user.token
    return requests.request(method,
                            '/'.join((obj.api_url, subpath)),
                            headers=headers,
                            **kwargs)


class GitLabEntity:
    """Abstract base class for various API entities (Project, Group…).

    Expected attributes or properties:

    - ``api_url``
    - ``owner_user``
    """

    def api_request(self, method, user=None, **kwargs):
        """Perform an API request as the given user.

        :param User user: if ``None``, defaults to the project owner

        Other params are passed down to :func:`api_request`.

        This method will eventually replace the older :meth:`api_request`,
        or be replaced by defaulting logic directly in the :func:`api_request`
        function.
        """
        if user is None:
            user = self.owner_user
        return api_request(method, self, user, **kwargs)

    # compat alias for subclass also having an `api_request` method
    user_api_request = api_request

    def api_get(self, **kwargs):
        """Perform a simple HTTP API GET, by default as the owner.

        :param kwargs: as in :meth:`user_api_request`
        """
        return self.user_api_request('GET', **kwargs)

    def api_put(self, **kwargs):
        """Perform a simple HTTP API PUT, by default as the owner.

        :param kwargs: as in :meth:`user_api_request`
        """
        return self.user_api_request('PUT', **kwargs)

    def api_post(self, **kwargs):
        """Perform a simple HTTP API POST, by default as the owner.

        :param kwargs: as in :meth:`user_api_request`
        """
        return self.user_api_request('POST', **kwargs)

    def api_delete(self, **kwargs):
        """Perform a simple HTTP API DELETE, by default as the owner.

        :param kwargs: as in :meth:`user_api_request`
        """
        return self.user_api_request('DELETE', **kwargs)



# Copyrightqhg 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import enum


class Access(enum.IntEnum):
    """Union of all possible values for Projects and Groups."""
    NO_ACCESS = 0
    MINIMAL = 5  # Introduced in GitLab 13.5
    GUEST = 10
    REPORTER = 20
    DEVELOPER = 30
    HG_PUBLISHER = 33  # Introduced in Heptapod 0.26
    MAINTAINER = 40
    OWNER = 50


class ProjectAccess(enum.IntEnum):
    """Subset of Access suitable to set on a Project.

    - `OWNER` is only settable on Groups.
    - `NO_ACCESS` is probably not settable on Projects
    - `MINIMAL` is to be determined
    """
    GUEST = Access.GUEST
    REPORTER = Access.REPORTER
    DEVELOPER = Access.DEVELOPER
    HG_PUBLISHER = Access.HG_PUBLISHER
    MAINTAINER = Access.MAINTAINER


class GroupAccess(enum.IntEnum):
    """Subset of Access suitable to set on a Group"""
    GUEST = Access.GUEST
    REPORTER = Access.REPORTER
    DEVELOPER = Access.DEVELOPER
    MAINTAINER = Access.MAINTAINER
    OWNER = Access.OWNER



# Copyright 2020-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import contextlib
import json
from io import BytesIO
import logging
import os
import requests
import selenium.webdriver
import shutil
import subprocess
import sys
import tarfile
import tempfile
import time
from urllib.parse import urlparse

from . import docker
from .constants import DEFAULT_RUN_DIR
from .namespace import Group
from .user import User

logger = logging.getLogger(__name__)

INITIAL_TIMEOUT = 600  # seconds

BRANCH_PROTECTIONS = dict(none=0,
                          dev_can_push=1,
                          full=2,
                          dev_can_merge=3,
                          )


class Heptapod:
    """Base class and minimum common control of Heptapod server.

    This is used directly in case `--heptapod-remote` is selected.
    """

    fs_access = False
    """True if manipulation of files is possible.

    Implies :attr:`repositories_root` not to be None.
    """

    shell_access = False
    """True if running arbitrary commands as in a shell is possible."""

    repositories_root = None
    """Path to the repositories, from the Heptapod server point of view.

    This is potentially not meaningful on the system that runs the test,
    see :class:`DockerHeptapod` for an example.
    """

    reverse_call_host = None
    """Network address for the system running these tests, seen from Heptapod.
    """

    webdriver_type = 'chrome'
    """Type of webdriver to use."""

    webdriver_remote_url = None
    """URL for a Selenium RemoteWebDriver."""

    wait_after_first_response = 0
    """Time to wait after we've had a first non error HTTP response.
    """

    hg_executable = 'hg'

    chrome_driver_args = ()

    firefox_driver_args = ()

    clever_cloud_sso = False
    """If ``True``, authentications will got through the Clever Cloud SSO.
    """

    default_user_name = 'root'
    """Default user for project, group creation etc."""

    default_group = None
    """Group instance where to create projects in by default."""

    instance_type = 'development'
    """The type of instance, meaning how these tests operate on it.

    It is of course completely possible to treat a development instance
    as if it were for production: that's what happens when developping the
    production server tests, of course.

    Treating a production instance as a development one is also technically
    possible (strongly discouraged of course).
    """

    def __init__(self, url,
                 ssh_user=None,
                 ssh_port=None,
                 ssh_keys_dir=None,
                 run_dir=DEFAULT_RUN_DIR,
                 hg_native=None,
                 clever_cloud_sso=False,
                 reverse_call_host=None,
                 wait_after_first_response=0,
                 webdriver_type=None,
                 webdriver_remote_url=None):
        self.parsed_url = urlparse(url)
        self.url = url
        self.ssh_port = ssh_port
        self.ssh_user = ssh_user
        self.ssh_keys_dir = ssh_keys_dir
        self.run_dir = run_dir
        self.hg_native = hg_native
        self.hg_without_git = hg_native == 'without-git'
        self.clever_cloud_sso = clever_cloud_sso
        self.users = {}
        if reverse_call_host is not None:
            self.reverse_call_host = reverse_call_host
        self.dead = None
        if webdriver_type is not None:
            self.webdriver_type = webdriver_type
        self.webdriver_remote_url = webdriver_remote_url
        self.wait_after_first_response = wait_after_first_response
        self.settings = {}

    @property
    def heptapod(self):
        return self

    @property
    def ssh_url(self):
        if self.ssh_user is None or self.ssh_port is None:
            raise ValueError("Missing configuration for SSH")
        return 'ssh://{user}@{host}:{port}'.format(
            host=self.host,
            user=self.ssh_user,
            port=self.ssh_port,
        )

    @property
    def host(self):
        return self.parsed_url.netloc.rsplit(':', 1)[0]

    @property
    def api_url(self):
        return '/'.join((self.url, 'api', 'v4'))

    @property
    def root_token_headers(self):
        return {'Private-Token': self.users['root'].token}

    @property
    def basic_user_token_headers(self):
        return {'Private-Token': self.users['test_basic'].token}

    def run_shell(self, command, **kw):
        exit_code, output = self.execute(command, **kw)
        if exit_code != 0:
            raise RuntimeError(
                ('Heptapod command {command} returned a non-zero '
                 'exit code {exit_code}').format(
                     command=command,
                     exit_code=exit_code,
                ))
        return output

    def get_user(self, name):
        """Return a :class:`User` instance, or `None`."""
        return self.users.get(name)

    def new_webdriver(self):
        if self.webdriver_type == 'chrome':
            cls = selenium.webdriver.Chrome
            options = selenium.webdriver.ChromeOptions()
            for arg in self.chrome_driver_args:
                options.add_argument(arg)
            options.add_argument('--headless')
        elif self.webdriver_type == 'firefox':
            cls = selenium.webdriver.Firefox
            options = selenium.webdriver.FirefoxOptions()
            for arg in self.firefox_driver_args:
                options.add_argument(arg)
            options.add_argument('--headless')

        if self.webdriver_remote_url:
            return selenium.webdriver.Remote(
                command_executor=self.webdriver_remote_url,
                options=options
            )
        return cls(options=options)

    def wait_startup(self, first_response_timeout=INITIAL_TIMEOUT,
                     wait_after_first_response=None):
        """Wait for Heptapod to be ready after startup.

        We have to take into account that the server may have just started
        (that's frequent in development and it's annoying for a human to
        wait) or could even be starting from scratch, configuring itself,
        creating the DB schema etc. (typical of CI).
        In that latter case, an amount of extra wait after the first successful
        HTTP response is often needed.
        """
        logger.info("Waiting for Heptapod to answer requests")
        dead_msg = ("Heptapod server did not "
                    "respond in %s seconds" % first_response_timeout)
        start = time.time()
        while time.time() < start + first_response_timeout:
            try:
                resp = requests.get(self.url, allow_redirects=False)
            except IOError:
                resp = None

            if resp is None:
                logger.debug("Couldn't reach Heptapod")
            elif resp.status_code >= 400:
                logger.debug("Heptapod response code %r", resp.status_code)
            else:
                logger.info("Heptapod is up")
                self.dead = False
                if wait_after_first_response:
                    logger.info("Waiting additional %d seconds "
                                "after first successful HTTP call",
                                wait_after_first_response)
                    time.sleep(wait_after_first_response)
                return

            duration = 1
            logger.debug("Retrying in %.1f seconds", duration)
            time.sleep(duration)

        self.dead = True
        raise AssertionError(dead_msg)

    def instance_cache_file(self):
        return self.run_dir / 'instance.cache'

    def load_instance_cache(self):
        path = self.instance_cache_file()

        def invalidate_retry():
            logger.warning("Removing cache file %r and starting afresh.", path)
            os.unlink(path)
            return self.load_instance_cache()

        try:
            with open(path) as cachef:
                cached = json.load(cachef)

        except Exception:
            logger.info("Cache file %r not available or not readable. "
                        "Heptapod instance info will be retrieved "
                        "or initialized", path)
        else:
            instance_type = cached.get('instance_type')
            if instance_type != self.instance_type:
                # None means before the introduction of this invalidation
                # in case anyone wonders (not worth a specific message)
                logger.warning(
                    "Cache file %r is for another instance type (%r) ",
                    path, instance_type)
                return invalidate_retry()

            url = cached.get('url')
            # for now all development instances have the same two
            # fixed test users (root and test_basic). There is already
            # a token invalidation logic.
            if instance_type == 'production' and url != self.url:
                # None means before the introduction of this invalidation
                # in case anyone wonders (not worth a specific message)
                logger.warning(
                    "Cache file %r is for another instance (%r) ",
                    path, url)
                return invalidate_retry()

            for name, info in cached['users'].items():
                if 'id' not in info:
                    logger.warning("Cache file %r is from an earlier version "
                                   "of heptapod-tests. ", path)
                    return invalidate_retry()
                self.users[name] = User(heptapod=self,
                                        name=name,
                                        id=info['id'],
                                        token=info['token'])

    def update_instance_cache(self):
        users = {user.name: dict(token=user.token, id=user.id)
                 for user in self.users.values()}
        with open(self.instance_cache_file(), 'w') as cachef:
            json.dump(dict(url=self.url,
                           instance_type=self.instance_type,
                           users=users), cachef)

    def prepare(self, root_password):
        """Make all preparations for the Heptapod instance to be testable.

        This currently amounts to

        - defining the root password
        - activating and retrieving a root API token
        - creating a persistent `test_basic` user
        - activating and retrieving an API token for `test_basic`
        - keeping a logged in webdriver for each persistent user
        """
        assert not self.dead, "Heptapod server marked dead by previous test."

        if self.dead is None:
            self.wait_startup(
                wait_after_first_response=self.wait_after_first_response)
        self.load_instance_cache()

        logger.info("Preparing root user.")
        start = time.time()
        root = self.users.get('root')
        if root is None:
            root = User.init_root(self, root_password)
        else:
            root.password = root_password
        root.ensure_private_token()

        logger.info("Preparing application settings.")
        wait_settings = self.set_application_settings(
            # necessary if we want to listen to web hooks from these tests
            # in GitLab v12.2.0, this is deprecated for ...from_web_hooks...
            allow_local_requests_from_web_hooks_and_services="true",
            # needed in GitLab ≥14.8
            api_cache_expiration_factor=0.0,
            # Depending on Heptapod version, native or non-native Mercurial
            # VCS type may not be allowed by default.
            vcs_types='hg,git,hg_git',
        )

        feature_flags = dict(
            api_caching_branches=False,
        )
        feature_flags['hg_without_git'] = self.hg_without_git
        wait_flags = self.set_feature_flags(feature_flags)

        ttw = max(wait_settings, wait_flags)
        if ttw:
            logger.warning("Application settings or Feature flags updated. "
                           "Waiting %d seconds for the change "
                           "to be visible by all server processes", ttw)
            time.sleep(ttw)

        logger.info("Preparing basic user.")
        User.ensure(self, 'test_basic',
                    fullname='Bäsîc Test',
                    password='test_basic')

        if self.ssh_keys_dir is not None:
            logger.info("Uploading users SSH keys.")
            self.load_ssh_keys()
            self.upload_ssh_pub_keys()
            subprocess.call((
                'ssh-keygen',
                '-R', '[{host}]:{port}'.format(
                    host=self.host, port=self.ssh_port)))
        self.update_instance_cache()
        logger.info("All preparations done in %.2f seconds. "
                    "Proceeding with tests.", time.time() - start)

    def set_application_settings(self, **settings):
        """Change GitLab application settings and update :attr:`settings`.

        :returns: the type to wait in seconds for cache expiration
        """
        settings_url = self.api_url + '/application/settings'
        existing = requests.get(settings_url,
                                headers=self.root_token_headers).json()
        # The format for update is different from the read format.
        # For instance, `vcs_types` is read as a list, but has to be set
        # as a comma-separated string. Therefore the simplest is not to
        # update only if needed, rather to update inconditionally and
        # check after the fact if anything was changed
        resp = requests.put(
            settings_url,
            headers=self.root_token_headers,
            data=settings,
        )
        assert resp.status_code == 200
        new_settings = resp.json()
        if existing == new_settings:
            logger.info("Application settings were already set as required.")
            return 0

        self.settings = new_settings
        # TODO the cache expiration is configurable as
        # Gitlab.config.gitlab['application_settings_cache_seconds']
        # (see app/models/concerns/cacheable_attributes.rb)
        logger.info("Application settings changed")
        return 60

    def sync_application_settings(self):
        """Update :attr:`settings` for current values"""
        resp = requests.get(
            self.api_url + '/application/settings',
            headers=self.root_token_headers,
        )
        assert resp.status_code == 200
        self.settings = resp.json()

    def set_feature_flags(self, flags):
        """Make sure that GitLab feature flags are as specified

        This controls instance-wide value of the feature flags. Any value
        subsequently set for a specific actor (Project, User etc) will
        be ignored.

        :param flags: `dict` mapping flag names to the wished boolean value.
        :returns: time to wait for cache expiration, in seconds

        Uses the dedicated API:
         https://docs.gitlab.com/ce/api/features.html#set-or-create-a-feature
        """
        if self.check_feature_flags(flags):
            logger.info("Feature flags %r already set as required",
                        flags)
            return 0

        headers = self.root_token_headers
        for flag, value in flags.items():
            resp = requests.post('/'.join((self.api_url,
                                           'features',
                                           flag)),
                                 headers=headers,
                                 # True doesn't seem to do anything, so
                                 # let's go the percentage way.
                                 data=dict(value=100 if value else 0)
                                 )
            assert resp.status_code < 400

        # Feature flags have several level of caching, with L1 being
        # a process-level RAM cache (memoization) with no invalidation
        # but expiration set to 1 minute (see lib/feature.rb in the Rails app
        # source).
        # We have no other choice than waiting:
        return 60

    @property
    def feature_flags_defaults(self):
        defaults = getattr(self, '_feature_flags_defaults', None)
        if defaults is None:
            resp = requests.get(self.api_url + '/features/definitions',
                                headers=self.root_token_headers)
            assert resp.status_code == 200
            self._feature_flags_defaults = defaults = {
                f['name']: f['default_enabled'] for f in resp.json()
            }

        return defaults

    def feature_flags_values(self):
        """Return boolean values of all defined feature flags.

        Like all feature flags treatment in this class, it does not
        understand partially activated flags (percentage etc.).

        The main API point for feature flags only lists the persisted
        values, i.e. does not take into account those that have never been
        mutated since they were defined (first installation or update that
        includes them), so we have to merge them with defaults.
        """
        flags = self.feature_flags_defaults.copy()
        resp = requests.get(self.api_url + '/features',
                            headers=self.root_token_headers)
        assert resp.status_code == 200
        for item in resp.json():
            logger.debug("Persisted feature flag response item: %r", item)
            flags[item['definition']['name']] = item['state'] == 'on'

        return flags

    def check_feature_flags(self, expected_flags):
        current_flags = self.feature_flags_values()
        for name, expected in expected_flags.items():
            current = current_flags.get(name)
            if current is None:
                raise RuntimeError("Feature flag %r is not defined in this "
                                   "Heptapod instance" % name)
            if current is not expected:
                logger.info("Value of feature flag %r is %s, differing from "
                            "the expected %s", name, current, expected)
                return False

        return True

    def api_request(self, method, user=None, subpath='', **kwargs):
        """Perform a simple API HTTP request

        `method` is the HTTP method to use, same as in `requests.request`.

        The full URL is made of the API URL of the instance, together with
        the given subpath (example 'snippets/42').

        Appropriate authentication headers are added on the fly.

        :param user: the :class:`User` to run the request as. If not specified,
                     the request is sent as the root user.

        All other kwargs are passed to `requests.request()`
        """
        headers = kwargs.pop('headers', {})
        token = self.owner_token if user is None else user.token
        headers['Private-Token'] = token
        return requests.request(method,
                                '/'.join((self.api_url, subpath)),
                                headers=headers,
                                **kwargs)

    def api_get(self, **kwargs):
        return self.api_request('GET', **kwargs)

    def api_post(self, **kwargs):
        return self.api_request('POST', **kwargs)

    def api_put(self, **kwargs):
        return self.api_request('PUT', **kwargs)

    def api_delete(self, **kwargs):
        return self.api_request('DELETE', **kwargs)

    def load_ssh_keys(self, key_name_mapping=None):
        """Load client-side information to use SSH keys

        Also makes sure the keys are actually usable (perms)

        :param key_name_mapping: allow to control the variable part of the
           key file name instead of simply using the user name.
        """
        if key_name_mapping is None:
            key_name_mapping = {}

        ssh_dir = self.ssh_keys_dir
        for name, user in self.users.items():
            base_fname = 'id_rsa_heptapod_' + key_name_mapping.get(name, name)
            priv = ssh_dir / base_fname
            pub = ssh_dir / (base_fname + '.pub')

            # VCSes tend not to preserve non-executable perm bits
            priv.chmod(0o600)
            user.ssh = dict(priv=str(priv), pub=pub.read_text())

    def upload_ssh_pub_keys(self):
        """Upload SSH public keys for all users to Heptapod."""
        for user in self.users.values():
            user.ensure_ssh_pub_key(user.ssh['pub'])

    def close_webdrivers(self):
        for user in self.users.values():
            user.close_webdrivers()

    def close(self):
        if self.dead is not False:
            return
        self.close_webdrivers()

    def execute(self, command, **kw):
        raise NotImplementedError('execute')

    def force_remove_route(self, route_path, source_type='Project'):
        logger.error("Attempt to force-remove route %r, not implemented "
                     "for %r", route_path, self.__class__)
        raise NotImplementedError('force_remove_route')

    def gitlab_ctl(self, command, services=None):
        """Apply service management command.

        'command' would typically be 'start', 'stop', etc.

        :param services: an iterable of service names (who can themselves
                         be different depending on the concrete subclass).
                         If supplied, the command will apply only to those
                         services.
        """
        raise NotImplementedError('gitlab_ctl')

    def rake(self, *args):
        """Call GitLab Rake"""
        raise NotImplementedError('rake')

    def psql(self, statement):
        """Calls psql on the instance database.

        This is *not* an access through the client library, we cannot separate
        arguments as is usually mandatory in regular applications to prevent
        SQL injections.
        """
        raise NotImplementedError('psql')

    def remove_all_backups(self):
        """Remove all existing backups with no impediment for new backups.
        """
        raise NotImplementedError('remove_all_backups')

    def backup_create(self, clean_previous=True):
        """Create a new backup

        :param bool clean_previous: if true, any previously existing backups
            are removed. This is handy so that the restore rake task knows
            which one to restore without any need to tell it.
        """
        if clean_previous:
            self.remove_all_backups()
        start = time.time()
        logger.info("Starting backup task")
        self.rake('gitlab:backup:create')
        elapsed = time.time() - start
        logger.info("Backup task finished in %d seconds (%.1f minutes)",
                    elapsed, elapsed / 60)

    @contextlib.contextmanager
    def backup_restore(self):
        """Context manager for backups restoration.

        This is a context manager as a way to provide resuming of the
        tests session on errors, in a convenient way for the caller.
        That means ensuring as much as possible that the server is running,
        maybe wait again for it, reinitialize passwords and tokens…
        """
        try:
            self.gitlab_ctl('stop', services=self.RAILS_SERVICES)
            start = time.time()
            logger.info("Starting backup restoration")
            self.rake('gitlab:backup:restore', 'force=yes')
            elapsed = time.time() - start
            logger.info("Backup restoration finished in %d seconds "
                        "(%.1f minutes). Now restarting Rails.",
                        elapsed, elapsed / 60)
            self.gitlab_ctl('start', services=self.RAILS_SERVICES)

            self.wait_startup()
            yield

        except Exception:
            logger.error("Backup restoration failed")
            # these are idempotent
            self.gitlab_ctl('start', services=self.RAILS_SERVICES)

            # Worst case scenario, we lost all our data. We need to
            # reprepare the server for subsequent tests
            self.prepare(self.users['root'].password)
            raise

    def set_vcs_types_settings(self, vcs_types):
        self.set_application_settings(vcs_types=','.join(vcs_types))

    def apply_hashed_storage_setting(self, hashed_storage):
        # TODO it would be tempting not to restart if the setting is already
        # at the wished value, but this currently cannot take into account
        # rollbacked values that aren't followed by a restart. This will
        # be more complicated and take more time than we can afford right now
        # to really make work.
        self.set_application_settings(hashed_storage_enabled=hashed_storage)

        # let's be sure that redis is restarted when the Rails services
        # start
        self.gitlab_ctl('stop', self.RAILS_SERVICES)
        self.gitlab_ctl('stop', ['redis'])
        # we restart everything in case a service would depend on Redis
        # and would fail to reconnect automatically

        # closing all webdrivers, because restart of Redis will kill sessions
        self.close_webdrivers()
        self.gitlab_ctl('restart')
        self.wait_startup()

        # recheck that the setting is applied
        self.sync_application_settings()
        assert self.settings['hashed_storage_enabled'] is hashed_storage


class ProductionHeptapod(Heptapod):
    """An Heptapod server for which we don't have the root password.

    This allows tests to be naturally sandboxed, typically in the groups
    and personal namespaces of a couple of users.
    """

    instance_type = 'production'

    def __init__(self, group_owner_credentials, default_group_id,
                 **kw):
        super(ProductionHeptapod, self).__init__(**kw)
        self.load_instance_cache()
        self.group_owner = self.init_user(group_owner_credentials)
        self.default_user_name = self.group_owner.name
        self.default_group_id = default_group_id

    def init_user(self, cli_credentials):
        """Instantiate a viable `User` from pytest options credentials."""
        user_id, name, email, pwd = cli_credentials.split(':', 3)
        user_id = int(user_id)
        user = self.users.get(name)
        if user is None:
            user = self.users[name] = User(heptapod=self,
                                           id=user_id,
                                           email=email,
                                           name=name,
                                           password=pwd)
        else:
            # id and password could have changed since last execution
            user.id = user_id
            user.password = pwd
            # email is not cached but can actually be necessary
            # (identifier for SSO login)
            user.email = email

        return user

    def prepare(self, root_password):
        """Prepare the group owner.

        root password is just ignored in this subclass.
        """
        start = time.time()
        assert not self.dead, "Heptapod server marked dead by previous test."

        if self.dead is None:
            self.wait_startup(
                wait_after_first_response=self.wait_after_first_response)

        logger.info("Preparing Group owner: %r", self.group_owner)
        self.group_owner.ensure_private_token()

        self.update_instance_cache()
        logger.info("All preparations done in %.2f seconds. "
                    "Proceeding with tests.", time.time() - start)

        self.load_ssh_keys(key_name_mapping={
            self.group_owner.name: 'group_owner',
        })
        self.upload_ssh_pub_keys()
        self.init_default_group()

    def init_default_group(self):
        gid = self.default_group_id
        if gid is None:
            groups = self.group_owner.api_owned_groups()
            min_depth = None
            for group in groups:
                depth = group['full_path'].count('/')
                if min_depth is None or depth < min_depth:
                    min_depth = depth
                    gid = group['id']

        if gid is None:
            raise RuntimeError(
                "Could not find a group owned by %r" % self.group_owner)
        self.default_group = Group.api_retrieve(
            self, gid, owner_name=self.group_owner.name)

    def remove_ssh_keys(self):
        for user in (self.group_owner, ):
            user.delete_ssh_keys()

    def close(self):
        super(ProductionHeptapod, self).close()
        self.remove_ssh_keys()


class OmnibusHeptapod(Heptapod):

    fs_access = True

    shell_access = True

    repositories_root = '/var/opt/gitlab/git-data/repositories'

    backups_dir = '/var/opt/gitlab/backups'

    RAILS_SERVICES = ('puma', 'sidekiq')

    gitlab_ctl_command = ('sudo', 'gitlab-ctl')

    reverse_call_host = '127.0.0.1'

    ssh_url = 'ssh://git@localhost'

    # TODO read from conf for logical dependency loosening.
    # Not an immediate priority, since we're not concerned about Python 2
    # any more (see heptapod#353)
    hg_executable = '/opt/gitlab/embedded/bin/hg'

    chrome_driver_args = ('--no-sandbox', )

    def execute(self, command, user='git'):
        if user != 'git':
            raise NotImplementedError(
                "On Omnibus Heptapod, only 'git' user is allowed")

        logger.debug("OmnibusHeptapod: executing command %r", command)
        process = subprocess.Popen(command,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        out, err = [o.decode() for o in process.communicate()]
        retcode = process.poll()

        if out:
            sys.stdout.write(out)
        if err:
            sys.stderr.write(err)
        return retcode, out

    def put_archive(self, dest, path, owner='git'):
        # this assumes owner has write access to parent directory,
        # but that should be enough, and avoids some processing as root
        self.run_shell(['tar', '-xf', path, '-C', dest], user=owner)

    def put_archive_bin(self, dest, fobj):
        with tempfile.NamedTemporaryFile() as tempf:
            tempf.write(fobj.read())
            self.put_archive(dest, tempf.name)

    def put_file_lines(self, path, lines, uid=998, gid=998):
        with open(path, 'w') as fobj:
            fobj.writelines(lines)
        os.chown(path, uid, gid)

    def path_exists(self, path):
        return os.path.exists(path)

    def get_file_lines(self, path):
        with open(path) as fobj:
            return fobj.readlines()

    def remove_all_backups(self):
        # using find in order not to rely on shell expansion for *.tar
        self.run_shell(('find', '/var/opt/gitlab/backups',
                        '-name', '*.tar',
                        '-delete'))

    def rake(self, *args):
        cmd = ['gitlab-rake']
        cmd.extend(args)
        code, out = self.execute(cmd, user='git')
        return out.encode()  # Consistency with HDK (returns bytes)

    def gitlab_ctl(self, command, services=None):
        base_cmd = self.gitlab_ctl_command + (command, )
        if services is None:
            self.run_shell(base_cmd)
        else:
            for service in services:
                self.run_shell(base_cmd + (service, ))

    def psql(self, statement):
        # db name is hardcoded.
        # This is good enough for now, but we could
        # read it from config if needed.
        return self.run_shell(('gitlab-psql', 'gitlabhq_production',
                               '-c', statement))


class DockerHeptapod(OmnibusHeptapod):

    gitlab_ctl_command = ('gitlab-ctl', )

    git_executable = 'git'

    chrome_driver_args = ()

    def __init__(self, docker_container, **kw):
        super(DockerHeptapod, self).__init__(**kw)
        self.docker_container = docker_container
        if self.reverse_call_host is None:
            self.reverse_call_host = docker.host_address(docker_container)

    def execute(self, command, user='root'):
        return docker.heptapod_exec(self.docker_container, command, user=user)

    def run_shell(self, command, **kw):
        return docker.heptapod_run_shell(self.docker_container, command, **kw)

    @property
    def ssh_url(self):
        return super(OmnibusHeptapod, self).ssh_url

    def put_archive(self, dest, path, owner='git'):
        res = docker.heptapod_put_archive(self.docker_container, dest, path)
        self.run_shell(['chown', '-R', 'git:root', dest])
        return res

    def put_archive_bin(self, dest, fobj):
        return docker.heptapod_put_archive_bin(
            self.docker_container, dest, fobj)

    def get_archive(self, path, tarf):
        return docker.heptapod_get_archive(self.docker_container, path, tarf)

    def path_exists(self, path):
        code, out = self.execute(['stat', path])
        return code == 0

    def put_file_lines(self, path, lines, uid=998, gid=998):
        dirpath, filename = path.rsplit('/', 1)
        tar_buf = BytesIO()
        tarf = tarfile.open(mode='w:', fileobj=tar_buf)

        tinfo = tarfile.TarInfo(name='hgrc')
        contents_buf = BytesIO()
        contents_buf.writelines(l.encode() for l in lines)
        tinfo.size = contents_buf.tell()
        tinfo.uid, tinfo.gid = uid, gid
        contents_buf.seek(0)
        tarf.addfile(tinfo, fileobj=contents_buf)

        tar_buf.seek(0)
        self.put_archive_bin(dirpath, tar_buf)

    def get_file_lines(self, path):
        dirname, filename = path.rsplit('/', 1)
        buf = BytesIO()
        self.get_archive(path, buf)
        buf.seek(0)
        tarf = tarfile.open(mode='r:', fileobj=buf)
        return [l.decode() for l in tarf.extractfile(filename).readlines()]

    def force_remove_route(self, route_path, source_type='Project'):
        """Delete a route from the database.

        Sometimes GitLab fails to clean Project routes after failed tests.
        """
        logger.warn("Cleaning up leftover route at %r", route_path)
        self.psql("DELETE FROM routes "
                  "WHERE source_type='%s' "
                  "  AND path='%s'" % (source_type, route_path))


class SourceHeptapod(Heptapod):
    """An Heptapod server installed from source on the same system.

    Same system means without using any container technology (Docker or
    otherwise) that would insulate the tests from the server.
    """

    fs_access = True

    shell_access = True

    reverse_call_host = '127.0.0.1'

    @property
    def ssh_url(self):
        return 'ssh://{host}:{port}'.format(
            host=self.host,
            port=self.ssh_port,
        )

    def __init__(self, repositories_root, **kw):
        super(SourceHeptapod, self).__init__(**kw)
        self.repositories_root = repositories_root

    def execute(self, command, user='git'):
        if user != 'git':
            raise NotImplementedError(
                "On source Heptapod, only same user as for Rails and HgServe "
                "is allowed")
        logger.debug("SourceHeptapod: executing command %r", command)
        process = subprocess.Popen(command,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        out, err = [o.decode() for o in process.communicate()]
        retcode = process.poll()

        if out:
            sys.stdout.write(out)
        if err:
            sys.stderr.write(err)
        return retcode, out

    def put_archive(self, dest, path, owner='git'):
        if owner != 'git':
            raise NotImplementedError(
                "On source Heptapod, only same owner as for Rails and HgServe "
                "is allowed")
        subprocess.check_call(['tar', 'xf', path], cwd=dest)

    def get_file_lines(self, path):
        with open(path, 'r') as fobj:
            return fobj.readlines()

    def put_file_lines(self, path, lines):
        with open(path, 'w') as fobj:
            fobj.writelines(lines)

    def path_exists(self, path):
        return os.path.exists(path)


class GdkHeptapod(SourceHeptapod):
    """An Heptapod server running with the GDK.
    """

    fs_access = True

    shell_access = True

    reverse_call_host = '127.0.0.1'

    RAILS_SERVICES = ('rails-web', 'rails-background-jobs')

    def __init__(self, gdk_root, **kw):
        self.gdk_root = gdk_root
        self.rails_root = os.path.join(gdk_root, 'gitlab')
        super(GdkHeptapod, self).__init__(
            repositories_root=os.path.join(gdk_root, 'repositories'),
            **kw)

    @property
    def backups_dir(self):
        return os.path.join(self.rails_root, 'tmp', 'backups')

    def remove_all_backups(self):
        if os.path.exists(self.backups_dir):
            shutil.rmtree(self.backups_dir)
        # as of GitLab 12.10, parent dir is always present
        os.mkdir(self.backups_dir)

    def rake(self, *args):
        cmd = ['bundle', 'exec', 'rake']
        cmd.extend(args)
        logger.debug("GdkHeptapod: calling %r", cmd)
        return subprocess.check_output(cmd, cwd=self.rails_root)

    def gitlab_ctl(self, command, services=None):
        base_cmd = ('gdk', command)

        def do_command(*opt_args):
            cmd = base_cmd + opt_args
            logger.debug("GdkHeptapod: calling %r", cmd)
            subprocess.check_call(cmd, cwd=self.rails_root)

        if services is None:
            do_command()
        else:
            for service in services:
                do_command(service)

    def psql(self, statement):
        # Assuming Unix socket, and db name is hardcoded.
        # This is good enough for now, but we could
        # read connection parameters from gdk.yml if needed.
        return self.run_shell(('psql',
                               '-h', os.path.join(self.gdk_root, 'postgresql'),
                               '-c', statement,
                               'gitlabhq_development',
                               ))



"""Handle network calls from Heptapod to the system running the tests.

This necessary to test such things as outbound web hooks and integrations.
"""
from multiprocessing import Process, Queue
from wsgiref.simple_server import make_server


def http_server(queue):
    """Create an HTTP server to receive calls from Heptapod.

    This is intended to run as a child process, with the
    The given `queue` been used to return information to the parent process.

    The first message is the bound port (as an integer)
    """
    def app(environ, start_response):
        status = '204 No Content'
        headers = [('Content-type', 'text/plain; charset=utf-8')]

        start_response(status, headers)
        meth = environ['REQUEST_METHOD']
        if meth in ('POST', 'PUT'):
            length = int(environ['CONTENT_LENGTH'])
            body = environ['wsgi.input'].read(length)
        else:
            body = None

        queue.put(dict(method=meth,
                       body=body,
                       environ={k: v for k, v in environ.items()
                                if isinstance(v, str)}))
        return ''

    with make_server('0.0.0.0', 0, app) as httpd:
        port = httpd.server_address[1]
        queue.put(port)
        httpd.handle_request()


class HttpListener():
    """An HTTP server to listen to HTTP requests from Heptapod.

    The response given by this server is always a `204 No Content`

    Adter init, `self.queue` will receive messages corresponding to
    the handled requests. Each is a `dict` with the following keys:

    - method: the HTTP method (POST, GET etc.)
    - environ: string-valued elements of the WSGI environment (in particular
      all HTTP headers should be there)
    - body: the request body (bytes), or `None` according to `method`
    """

    def __init__(self, heptapod):
        self.queue = Queue()
        self.process = Process(target=http_server,
                               args=(self.queue,))
        self.process.start()
        self.port = self.queue.get()
        self.heptapod = heptapod

    def join(self, timeout):
        self.process.join(timeout)
        if self.process.exitcode is None:  # we've reached the timeout
            self.process.terminate()
        return self.process.exitcode

    def url(self, heptapod):
        """The URL as seen from Heptapod."""
        host = heptapod.reverse_call_host
        if ':' in host:  # should happen only with IPv6 address
            host = '[%s]' % host
        return 'http://%s:%d' % (host, self.port)



def gitlab_branch(branch='default', topic=None):
    """Return the GitLab branch for a branch/topic combination."""
    if topic is None:
        return 'branch/' + branch
    return '/'.join(('topic', branch, topic))



# Copyright 2018 Paul Morelle <madprog@htkc.org>
# Copyright 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import logging
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
import time

from .selenium import (
    wait_could_click_element,
    wait_element_visible,
    webdriver_wait_get,
)


logger = logging.getLogger(__name__)


def initialize_root(driver, heptapod, password):
    # Create initial password
    elem = driver.find_element(By.NAME, 'user[password]')
    elem.send_keys(password)
    elem = driver.find_element(By.NAME, 'user[password_confirmation]')
    elem.send_keys(password)
    elem.send_keys(Keys.RETURN)

    sign_in_page_login(driver, heptapod, 'root', password=password)


def finalize_password(driver, heptapod, user, password, current_password=None):
    """Enter the final password, if already on the "Set up new password" page.
    """
    if current_password is None:
        current_password = password

    elem = driver.find_element(By.NAME, 'user[password]')
    elem.send_keys(current_password)
    elem = driver.find_element(By.NAME, 'user[new_password]')
    elem.send_keys(password)
    elem = driver.find_element(By.NAME, 'user[password_confirmation]')
    elem.send_keys(password)
    elem.send_keys(Keys.RETURN)


def sign_in_page_login(driver, heptapod, user_name, password=None):
    """Perform login as user, with webdriver already on the signin page.

    If password is not specified, it is read from the heptapod object.
    If it is, it'll be stored in the heptapod object
    """
    user = heptapod.users[user_name]
    if password is None:
        password = user.password
    else:
        user.password = password
    if heptapod.clever_cloud_sso:
        user.vcs_token_only = True
        return sign_in_page_clever_cloud_login(driver, user)

    elem = driver.find_element(By.NAME, 'user[login]')
    elem.send_keys(user_name)
    elem = driver.find_element(By.NAME, 'user[password]')
    elem.send_keys(password)
    elem.send_keys(Keys.RETURN)

    # Wait for login to be complete
    wait_element_visible(driver,
                         By.XPATH,
                         '//li[@data-qa-selector="user_menu"]')


def sign_in_page_clever_cloud_login(driver, user):
    wait_could_click_element(
        driver,
        lambda d: d.find_element(By.ID, 'oauth-login-saml'))
    assert 'clever-cloud.com' in driver.current_url

    elem = driver.find_element(By.ID, 'login-email')
    elem.send_keys(user.email)
    elem = driver.find_element(By.ID, 'login-pwd')
    elem.send_keys(user.password)
    elem.send_keys(Keys.RETURN)

    assert driver.current_url.startswith(user.heptapod.url)


def page_is_password_setup(driver):
    """Return if the current page is the password setup page.

    This happens for instance after first login with a password considered
    to be temporary
    """
    # plural form to get an empty list instead of an exception
    return bool(driver.find_elements(By.NAME, 'user[password_confirmation]'))


def login_as_root(driver, heptapod, password):
    start = time.time()
    webdriver_wait_get(heptapod, driver, relative_uri='/users/sign_in')
    assert 'GitLab' in driver.title
    html = driver.find_element(By.TAG_NAME, 'html').get_attribute('innerHTML')

    if 'Please create a password for your new account.' in html:
        initialize_root(driver, heptapod, password)
    else:
        sign_in_page_login(driver, heptapod, 'root', password)
        if page_is_password_setup(driver):
            finalize_password(driver, heptapod, 'root', password)
            sign_in_page_login(driver, heptapod, 'root', password)
        logger.info("Made signed-in webdriver for user %r in %.2f seconds",
                    'root', time.time() - start)



import logging
import os
import subprocess

GIT_EXECUTABLE = 'git'

logger = logging.getLogger(__name__)


def git_call(cwd, cmd,
             env=None,
             ssh_cmd=None,
             check_return_code=True,
             expected_return_code=0):
    """Wrapper to run git, allowing process error or not.


    :param check_status_code:
        if ``True``, and hg status code isn't 0, raises like an stderr variant
        of ``subprocess.check_output()``.
    :param encoding: if `None`, the default encoding is used
    """
    gitcmd = [GIT_EXECUTABLE]
    gitcmd.extend(str(s) if not isinstance(s, bytes) else s for s in cmd)
    if env is None:
        env = os.environ.copy()

    if ssh_cmd is not None:
        env['GIT_SSH_COMMAND'] = ssh_cmd

    proc = subprocess.Popen(gitcmd,
                            cwd=str(cwd),
                            env=env,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    out, err = [b.decode('utf-8', 'replace') for b in proc.communicate()]
    retcode = proc.poll()
    if check_return_code and retcode != expected_return_code:
        # logging at ERROR level because if it had been expected
        # call would have used `check_return_code=False`
        logger.error("Failed git_call (code %d, expected %d), "
                     "stdout=%r, stderr=%r",
                     retcode, expected_return_code, out, err)
        raise subprocess.CalledProcessError(retcode, cmd, output=err)
    return retcode, out, err


class LocalRepo(object):
    """Represent a repository on the system where the tests are run."""

    def __init__(self, path):
        self.path = path

    @classmethod
    def init(cls, path, default_url=None):
        """Create a repository at the given path and attach it to the given URL.

        :param path: a pytest Path instance
        :param url: Optional Heptapod URL of the repo, will become the
                    default path for push/pull operations.
                    In most cases, one would want the URL to enclose
                    Basic Authentication credentials, but it's not necessary.


        """
        git_call('.', ['init', path])
        repo = cls(path)

        if default_url is not None:
            repo.git('remote', 'add', 'origin', default_url)

        repo.configure_default_user()
        return repo

    @classmethod
    def clone(cls, url, path, ssh_cmd=None, expected_return_code=0):
        """Clone the repository from given URL.

        :param path: a pytest Path instance
        """
        git_call('.', ['clone', url, path], ssh_cmd=ssh_cmd,
                 expected_return_code=expected_return_code)
        repo = cls.init(path)
        return repo

    def configure_default_user(self):
        self.git('config', 'user.name', 'Test Git')
        self.git('config', 'user.email', 'testgit@heptapod.test')

    def git(self, *args, **kwargs):
        """Invoke the `git` executable and return its stdout.

        This behaves as `subprocess.check_output` does.
        """
        return git_call(self.path, args, **kwargs)[1]

    def git_unchecked(self, *args, **kwargs):
        """Invoke the `git` executable without raising based on return code

        :returns: (process return code, stdout, stderr).
        """
        return git_call(self.path, args, check_return_code=False, **kwargs)

    def graphlog(self):
        return self.git('log', '--all', '--graph')

    def sha(self, ref):
        return self.git('log', '--pretty=%H', '-n1', ref).strip()



# Copyright 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shared Selenium utils."""

import logging
import time
from selenium.common.exceptions import (
    ElementNotVisibleException,
    ElementNotInteractableException,
    NoSuchElementException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait

from heptapod_tests.wait import (
    BASE_TIMEOUT,
    wait_assert,
)

logger = logging.getLogger(__name__)


def page_means_not_ready(driver):
    """Detect a few symptoms of Heptapod not fully ready"""
    if '502' in driver.title:
        return True
    if 'Gitlab::Webpack::Manifest::ManifestLoadError' in driver.title:
        return True
    return False


def raw_page_content(driver):
    """Extract the content of a page meant for GitLab raw downloads.

    The webdriver rewraps it into a HTML ``pre`` element.
    """
    try:
        return driver.find_element(By.TAG_NAME, 'pre').text
    except NoSuchElementException:
        raise AssertionError("Page is empty or not a GitLab raw page")


def webdriver_wait_get(heptapod, driver,
                       relative_uri='', timeout=300, retry_delay=1):
    """Get the given URL when it's ready.

    At this stage, we already got the 302 on the site base URL, it's not
    clear why we can still get 502s, maybe just because that's harder
    that the simple redirection. Anyway, that justifies a shorter timeout
    by default than the initial one.
    """
    url = heptapod.url + relative_uri
    start = time.time()
    dead_msg = ("Heptapod server did not give a successful response"
                "in %s seconds" % timeout)
    while True:
        if time.time() > start + timeout:
            heptapod.dead = True  # will abort subsequent tests
        driver.get(url)
        if not page_means_not_ready(driver):
            break
        logger.debug("Title of page at %r is %r, retrying in %.1f seconds",
                     url, driver.title, retry_delay)
        time.sleep(retry_delay)
    assert not heptapod.dead, dead_msg


def could_click_element(selector):
    def try_click_element(driver):
        try:
            elem = selector(driver)
            elem.click()
            return True
        except (ElementNotVisibleException,
                NoSuchElementException,
                ElementNotInteractableException):
            return False
        except Exception:
            import traceback
            traceback.print_exc()
            return False
    return try_click_element


def webdriver_wait(driver, timeout_factor=1):
    return WebDriverWait(driver, BASE_TIMEOUT * timeout_factor)


def wait_could_click_element(driver, selector, **wait_kw):
    webdriver_wait(driver, **wait_kw).until(could_click_element(selector))


def wait_element_visible(driver, loc_by, loc_expr, **wait_kw):
    return webdriver_wait(driver, **wait_kw).until(
        ec.visibility_of_element_located((loc_by, loc_expr))
    )


def wait_assert_in_page_source(driver, s):
    """Wait until the given string is in page source.

    Avoid doubts with eager evaluation of `driver.page_source` (would not
    happen in a lambda, but still).
    """
    wait_assert(lambda: driver.page_source,
                lambda source: s in source)


def assert_webdriver_not_error(webdriver):
    """assert that the current page is not GitLab's error page rendering.

    We don't have a reliable way to detect an error
    The page title would be just 500 in production and the error class name in
    # development mode.
    """
    title = webdriver.title
    assert "Error" not in title
    assert not any(("(%d)" % code) in title
                   for code in (500, 502, 504, 403, 404))



# Copyright 2020-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from contextlib import contextmanager
import time

from .namespace import Group
from .project import Project
from .utils import unique_name


@contextmanager
def project_fixture(heptapod, name_prefix, owner=None, group=None, **opts):
    # TODO provide importable from heptapod_tests
    if owner is None:
        owner = heptapod.default_user_name
    if group is None:
        group = heptapod.default_group
    name = '%s_%s' % (name_prefix, str(time.time()).replace('.', '_'))
    with Project.api_create(heptapod, owner, name,
                            group=group, **opts) as proj:
        yield proj


@contextmanager
def group_fixture(heptapod, path_prefix, creator_name='root', parent=None):
    with Group.api_create(heptapod, unique_name(path_prefix),
                          user_name=creator_name,
                          parent=parent) as group:
        yield group



# Copyright 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections import namedtuple
import json
import logging
import os
import subprocess

from .utils import (
    assert_message,
)
from .wait import (
    wait_assert,
)

SHORT_TEMPLATE = (
    '{rev}:{node|short} {desc|firstline} ({phase})\n'
    '{if (obsfate, "  obsolete:{obsfate}\n\n")}'
)
HG_EXECUTABLE = 'hg'
_hg_client_version = None  # a write-once constant

logger = logging.getLogger(__name__)


def hg_call(cwd, cmd, check_return_code=True, expected_return_code=0,
            encoding=None):
    """Wrapper to run hg, allowing process error or not.


    :param check_status_code:
        if ``True``, and hg status code isn't 0, raises like an stderr variant
        of ``subprocess.check_output()``.
    :param encoding: if `None`, the default encoding is used
    """
    hgcmd = [HG_EXECUTABLE]
    hgcmd.extend(str(s) if not isinstance(s, bytes) else s for s in cmd)
    if encoding is None:
        env = os.environ
    else:
        env = dict(os.environ)
        env['HGENCODING'] = encoding
    proc = subprocess.Popen(hgcmd,
                            cwd=str(cwd),  # -R not enough for hg init
                            env=env,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    out, err = [b.decode('utf-8', 'replace') for b in proc.communicate()]
    retcode = proc.poll()
    if check_return_code and retcode != expected_return_code:
        # logging at ERROR level because if it had been expected
        # call would have used `check_return_code=False`
        logger.error("Failed hg_call (code %d, expected %d), "
                     "stdout=%r, stderr=%r",
                     retcode, expected_return_code, out, err)
        raise subprocess.CalledProcessError(retcode, cmd, output=err)
    return retcode, out, err


def hg_client_version():
    global _hg_client_version

    if _hg_client_version is None:
        code, out, err = hg_call('.', ['debuginstall', '-T', '{hgver}'],
                                 check_return_code=False)
        assert code in (0, 1), err
        _hg_client_version = tuple(int(s) for s in out.split('.'))
        logger.info("Mercurial client executable %r, "
                    "version is %r", HG_EXECUTABLE, _hg_client_version)
    return _hg_client_version


class LocalRepo(object):
    """Represent a repository on the system where the tests are run."""

    hgrc_extensions_lines = ("[extensions]",
                             "evolve =",
                             "topic = ",
                             "rebase =",
                             "[ui]",
                             "username = Heptapod Tests <fonct@heptapod.test>"
                             )

    def __init__(self, path):
        self.path = path

    @classmethod
    def init(cls, path, default_url=None):
        """Create a repository at the given path and attach it to the given URL.

        :param path: a pytest Path instance
        :param url: Optional Heptapod URL of the repo, will become the
                    default path for push/pull operations.
                    In most cases, one would want the URL to enclose
                    Basic Authentication credentials, but it's not necessary.


        """
        hg_call('.', ['init', path])
        repo = cls(path)

        hgrc_lines = list(cls.hgrc_extensions_lines)
        if default_url is not None:
            hgrc_lines.extend(("", "[paths]", "default = " + str(default_url)))
        repo.hgrc_append_lines(hgrc_lines)

        return repo

    @classmethod
    def clone(cls, url, path):
        """Clone the repository from given URL.

        :param path: a pytest Path instance
        """
        repo = cls.init(path, default_url=url)
        repo.hg('pull', '-u')
        return repo

    def hgrc_append_lines(self, lines):
        with self.path.join('.hg', 'hgrc').open('a') as fp:
            fp.writelines(l + '\n' for l in lines)

    def hg(self, *args, **kwargs):
        """Invoke the `hg` executable and return its stdout.

        This behaves as `subprocess.check_output` does.
        """
        return hg_call(self.path, args, **kwargs)[1]

    def hg_unchecked(self, *args, **kwargs):
        """Invoke the `hg` executable without raising if return code is nonzero.

        :returns: (process return code, stdout, stderr).
        """
        return hg_call(self.path, args, check_return_code=False, **kwargs)

    def assert_hg_failure(self, *args,
                          error_message_expectations='',
                          stdout_expectations='',
                          stderr_expectations='',
                          **kwargs):
        """Invoke the ``hg` executable, expect and assert it to fail

        All ``expectations`` kwargs can be either `str` instances, compiled
        regular expressions or iterables of such. It will be asserted that the
        actual error output matches them.

        :param error_message_expectations: check done on the expected channel
          (stdout or stderr), which can depend on the Mercurial client
          version.
        :param stdout_expectations: check done unconditionally on stdout.
        :param stderr_expectations: check done unconditionally on stderr.
        :returns: (process return code, stdout, stderr).
        """
        code, out, err = self.hg_unchecked(*args, **kwargs)
        # code 1 is hg standard for "no change"
        assert code not in (0, 1)

        err_msg_chan = out if hg_client_version() < (5, 9) else err
        assert_message(err_msg_chan, error_message_expectations)

        assert_message(out, stdout_expectations)
        assert_message(err, stderr_expectations)

        return code, out, err

    def graphlog(self, hidden=False):
        """Return a full log.

        This is useful to print (captured by the runner and displayed if there
        are failures), maybe to send to a file.

        :param hidden: if True, also display hidden (obsolete) changesets.

        """
        cmd = ['log', '--graph', '-T', SHORT_TEMPLATE]
        if hidden:
            cmd.append('--hidden')
        return self.hg(*cmd)

    def changeset_extracts(self, template_exprs,
                           collection=tuple,
                           revs=None,
                           val_sep='|:|', eol="\n"):
        """Extract information from changesets, as a collection of entries.

        This does not represent the graph information, but the ordering given
        by `hg log`, on which this method is based, is preserved if using
        an ordered collection.

        :param template_exprs: an iterable of template keywords or pairs
           (attribute, template expression).

           Examples:
           - ``['rev', 'phase', 'desc']``
           - ``['phase', ('desc', 'desc|firstline')]``

           The second one will return named tuples with ``phase`` and ``desc``
           attributes.
        :param collection: the wished collection class. Must be instantiable
           by a generator expression, like `set` and `tuple` are.
        :return: a named tuple class whose attributes are the elements of
          ``template_exprs`` and a set of instances of that class,
          one for each changeset.

        WARNING: there is no escaping. The command will fail utterly
        if the rendering of one template expression contains ``val_sep`` or
`       ``eol``.
        """
        template_exprs = [
            (expr, expr) if isinstance(expr, str) else tuple(expr)
            for expr in template_exprs
        ]
        template = val_sep.join("{%s}" % expr[1] for expr in template_exprs)
        cmd = ['log', '-T', template + eol]
        if revs is not None:
            cmd.extend(('-r', revs))

        Extract = namedtuple('Extract', (expr[0] for expr in template_exprs))
        return Extract, collection(Extract(*line.split(val_sep))
                                   for line in self.hg(*cmd).split(eol)
                                   if line)

    def branches(self):
        """Return `hg branches` as a set of names."""
        return set(self.hg('branches', '-T', '{branch}\n').splitlines())

    def node(self, revspec):
        """Return the node id for specified revision (can be symbolic)"""
        return self.hg('log', '-T', "{node}", '-r', revspec)

    def init_gitlab_ci(self,
                       message="Init GitLab CI",
                       script=("grep foo foo",),
                       ):
        """Create an commit a .gitlab-ci.yml file.

        A convenience method to have a valid configuration without caring
        too much with what it entails.
        """
        ci_config = dict(job=dict(script=script))
        # JSON is a subset of YaML and part of Python standard library
        self.path.join('.gitlab-ci.yml').write(json.dumps(ci_config))
        self.hg('add', '.gitlab-ci.yml')
        self.hg('commit', '-m', message)

    def count_changesets(self):
        """Return the total number of changesets in repo.

        Obsolete changesets are taken into account.
        """
        # for empty repo, this gives 1 + (-1), hence the expected 0
        return 1 + int(self.hg('log', '-r', 'tip', '-T', '{rev}'))

    def wait_pull_new_changesets(self, number, *flags,
                                 retry_wait_factor=0.1, **kwargs):
        """Pull repeatedly until new changesets are retrieved by pulls.

        If not enough changesets are retrieved, this is an ``AssertionError``.

        :param int number: the minimal number of new changesets to pull
        :param flags: additional arguments to pass to the pull.
           Example::
             repo.wait_pull_changesets(2, '-u')
        :param kwargs: other :func:`wait_assert` keyword arguments.
        :raises: AssertionError if not enough changesets are retrieved before
           timeout.
        """
        # can be -1 (empty repo, and that is still consistent)
        expected = self.count_changesets() + number
        wait_assert(
            lambda: self.hg('pull', *flags),
            lambda _: self.count_changesets() >= expected,
            retry_wait_factor=retry_wait_factor,
            **kwargs,
        )


def assert_matching_changesets(repo1, repo2, fields, ordered=True, revs=None):
    """Assert changesets to have same information in the two given repos

    Of course all structural information of the changesets can be checked
    with just ``node``, but there is also outer information, like for example
    ``phases``, and ``bookmarks``, whose inclusion depends on the test case.

    Also, adding some human-readable information such as `desc` or `topic`
    can make matter more obvious to people engaged in debugging.

    The ordering can be kept or not, according to the test case.
    """
    collection = tuple if ordered else set
    extract1, extract2 = [
        repo.changeset_extracts(fields, collection=collection, revs=revs)[1]
        for repo in (repo1, repo2)
    ]
    assert extract1 == extract2


def cli_pushvars(pushvars):
    """Turn pushvars into CLI arguments

    :params dict pushvars: the wished push variables and their values both
    as strings, except the special case of ``True``, which is meant for
    actions (currently translates as, e.g, ``ci.skip=``, same as an empty
    string, but that can change).

    :return: an iterable of arguments that can simply appended to a
      :meth:`LocalRepo.hg` call.
    """
    cli_args = []
    for k, v in pushvars.items():
        if v is True:
            v = ''
        cli_args.extend(('--pushvars', '='.join((k, v))))
    return cli_args



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from pathlib import Path
import yaml


class HeptapodConfig:
    """Read Heptapod-specific configuration from its own file.

    The whole point it to use a different file than pytest, so that it
    can be easily ignored.
    """

    schema = {}  # option name to type, defaults etc.

    def __init__(self, pytest_conf):
        self.pytest_conf = pytest_conf
        self.file_conf = {}

    @classmethod
    def load(cls, pytest_conf):
        hpd_conf = cls(pytest_conf)
        rel_path = pytest_conf.getoption('heptapod_config')
        # as of this writing, pytest's rootdir is a py.path object,
        # does not interpret `/` as expected if rel_path is in fact absolute
        conf_path = Path(pytest_conf.rootdir) / rel_path
        if conf_path.exists():
            with conf_path.open() as conf_fobj:
                hpd_conf.file_conf = yaml.safe_load(conf_fobj)
        return hpd_conf

    def get(self, attr_name):
        sch_entry = self.schema.get(attr_name)
        if sch_entry is None:
            raise KeyError(f"Config item {attr_name} is undefined")

        action = sch_entry.get('action')
        # In case of `store_true` and `store_false`,
        # pytest config will normalize as bool
        # hence won't return ``None`` if not set on the CLI
        if action == 'store_true':
            pytest_notset = False
            action_bool = True
        elif action == 'store_false':
            pytest_notset = True
            action_bool = True
        else:
            pytest_notset = None
            action_bool = False

        pyv = self.pytest_conf.getoption(attr_name)
        if pyv is not pytest_notset:
            return pyv

        fv = self.file_conf.get(attr_name)
        if fv is None:
            return sch_entry.get('default')

        # Not exactly the same as setting validator to `bool`
        # Given the wide range of boolean evaluation, we prefer strict
        # type checking here.
        if action_bool and not isinstance(fv, bool):
            raise ValueError("In configuration file, only boolean values "
                             "are allowed for %r" % attr_name)

        validator = sch_entry.get('type')
        return fv if validator is None else validator(fv)

    @classmethod
    def cli_add_config_path_option(cls, pytest_parser):
        pytest_parser.addoption(
            '--heptapod-config',
            default='heptapod-tests.yml',
            help="Path to the Heptapod Tests YaML configuration file, in "
            "which all the other custoom option of these tests can be set")

    @classmethod
    def add_option(cls, pytest_parser, name, **kwargs):
        if not name.startswith('--'):
            raise ValueError(
                "Configuration options must be defined in CLI"
                "form (--something)")
        pytest_kw = kwargs.copy()
        # we don't want argparse to apply defaults before we pass over
        # to the configuration file.
        pytest_kw.pop('default', None)
        pytest_parser.addoption(name, **pytest_kw)

        attr_name = name[2:].replace('-', '_')
        cls.schema[attr_name] = kwargs.copy()



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import attr
from urllib.parse import (
    urlencode,
)

from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

from .api import api_request
from .selenium import (
    wait_could_click_element,
    webdriver_wait,
)
from .wait import (
    wait_assert,
)


@attr.s
class MergeRequest:
    project = attr.ib()
    iid = attr.ib()

    @property
    def api_url(self):
        return '%s/merge_requests/%d' % (self.project.api_url, self.iid)

    @property
    def url(self):
        return '%s/-/merge_requests/%d' % (self.project.url, self.iid)

    @classmethod
    def api_create(cls, source_project, source_branch,
                   target_branch='branch/default',
                   target_project=None,
                   user=None,
                   title="Created through API"):
        data = dict(id=source_project.id,
                    source_branch=source_branch,
                    target_branch=target_branch,
                    title=title)
        if target_project is None:
            target_project = source_project
        else:
            data['target_project_id'] = target_project.id
        if user is None:
            user = source_project.owner_user
        resp = api_request('post', source_project, user,
                           subpath='merge_requests',
                           data=data)

        assert resp.status_code in (200, 201)
        return cls(target_project, resp.json()['iid'])

    @classmethod
    def webdriver_create(cls, source_project, source_branch,
                         target_branch='branch/default',
                         target_project=None,
                         user=None):
        """Create a merge request through the Web UI and return it.
        """
        if user is None:
            user = source_project.owner_user
        if target_project is None:
            target_project = source_project

        driver = user.webdriver
        compare_qs = {
            'merge_request[source_project_id]': source_project.id,
            'merge_request[source_branch]': source_branch,
            'merge_request[target_project_id]': target_project.id,
            'merge_request[target_branch]': target_branch,
        }
        driver.get('{url}/merge_requests/new?{qs}'.format(
            url=source_project.url,
            qs=urlencode(compare_qs),
        ))
        assert 'new merge request' in driver.title.lower()
        assert source_project.name in driver.title

        wait_could_click_element(driver, lambda d: d.find_element(
            By.XPATH,
            '//input[@type="submit" '
            'and @value="Create merge request"]')
        )

        split_url = driver.current_url.rsplit('/', 2)
        assert split_url[-2] == 'merge_requests'
        # same as with API creations: the MR belongs to target Project
        return cls(target_project, int(split_url[-1]))

    def api_get(self, user=None, assert_existence=True, check=True, **params):
        if user is None:
            user = self.project.owner_user
        resp = api_request('get', self, user, params=params)
        if resp.status_code == 404 and not assert_existence:
            return None

        if not check:
            return resp

        assert resp.status_code < 400
        return resp.json()

    def wait_assert(self, condition,
                    with_rebase_progress=False,
                    with_diverged_count=False,
                    msg=None,
                    **kw):
        if msg is None:
            msg = (
                "The given condition on Merge Request %s was still not "
                "fulfilled after retrying for {timeout} seconds") % self
        mr_opts = dict(include_diverged_commits_count=with_diverged_count,
                       include_rebase_in_progress=with_rebase_progress)
        return wait_assert(
            lambda: self.api_get(assert_existence=False, **mr_opts),
            condition,
            msg=msg, **kw)

    def api_put(self, data, user=None, check=True, subpath=''):
        if user is None:
            user = self.project.owner_user

        resp = api_request('put', self, user, subpath=subpath, data=data)
        if not check:
            return resp

        assert resp.status_code < 400
        return resp.json()

    def api_edit(self, user=None, check=True, **data):
        return self.api_put(data, user=user, check=check)

    def api_rebase(self, user=None, **kw):
        """Trigger rebase and wait_assert for completion.

        :param kwargs: passed down to :meth:`wait_assert`
        """
        resp = self.api_put(None, user=user, subpath='rebase', check=False)
        assert resp.status_code == 202  # Means "accepted", this is async
        mr_info = self.wait_assert(lambda info: not info['rebase_in_progress'],
                                   with_rebase_progress=True)
        assert mr_info['merge_error'] is None

    def api_accept(self,
                   user=None,
                   wait_mergeability=True,
                   timeout_factor=1,
                   check_merged=True,
                   **opts):
        if wait_mergeability:
            self.wait_assert(
                lambda info: info.get('merge_status') == 'can_be_merged',
                timeout_factor=timeout_factor,
                msg="Mergeability wrong or still unknown after "
                "{timeout} seconds")

        resp = self.api_put(dict(merge_request_iid=self.iid,
                                 id=self.project.id,
                                 **opts),
                            subpath='merge',
                            user=user,
                            check=False)

        if not check_merged:
            return resp

        assert resp.status_code == 200
        info = resp.json()
        assert info['state'] == 'merged'
        return info

    def wait_assert_merged(self,
                           expected_source_branch=None,
                           expected_source_branch_category=None,
                           check_source_branch_removal=True,
                           **wait_kw):
        """Checks that after a while, state is `merged` plus options

        :param expected_source_branch: if passed, will check that once merged
            the source branch is the given one. Not taken from MR info
            before waiting, for it could have been damaged by a bug (e.g. in
            `PostReceiveWorker`.
        :param expected_source_branch_category: similar to
            ``expected_source_branch`` but only checks for correct category,
            where category is translated to prefix and slash (e.g., ``topic/``)
        :param check_source_branch_removal: if `True`, check that the
            source branch does not exist any more server-side.
        """
        info = self.wait_assert(lambda info: info.get('state') == 'merged',
                                msg="Merge not detected in {timeout} seconds",
                                **wait_kw)

        project = self.project

        # Web UI for the merge request is not broken
        webdriver = project.owner_webdriver
        webdriver.get(self.url)
        assert '(500)' not in webdriver.title

        source_branch = info['source_branch']
        if expected_source_branch is not None:
            assert source_branch == expected_source_branch

        if expected_source_branch_category:
            assert source_branch.startswith(
                expected_source_branch_category + '/')

        if check_source_branch_removal:
            project.wait_assert_api_branches(
                lambda branches: source_branch not in branches)

    def webdriver_get_commit_links(self):
        """Retrieve the commit links from the 'commits' panel of the MR."""
        webdriver = self.project.owner_webdriver
        # this is the 'commits' pane, loading is dynamic, hence the wait
        webdriver.get(self.url + '/commits')
        webdriver_wait(webdriver).until(
            EC.presence_of_all_elements_located(
                (By.CSS_SELECTOR, 'a.commit-row-message')))

        return webdriver.find_elements(By.CSS_SELECTOR, 'li.commit')

    def assert_user_cannot_merge(self, user):
        # let's decrease chances of false negatives by testing that user has
        # access to the MR (projects tend to be private by default).
        # We could do better by, e.g., adding a comment but that requires only
        # the Reporter role.
        resp = api_request('GET', self, user)
        assert resp.status_code == 200

        resp = self.api_accept(user=user, check_merged=False)
        assert resp.status_code >= 400

    def assert_commit_link(self, expected_text, expected_sha):
        """Assert that a commit link with expected text and sha is present.
        """
        for li in self.webdriver_get_commit_links():
            if expected_text in li.text:
                # finer assertions about commit hash widget
                copy_label = li.find_element(
                    By.XPATH,
                    # div with class 'label', in a div having a button child
                    './/div[button]/div[contains(@class, "label")]')
                copy_button = li.find_element(
                    By.XPATH,
                    './/button[@title="Copy commit SHA"]')
                assert copy_button.get_attribute(
                    'data-clipboard-text') == expected_sha
                assert expected_sha.startswith(copy_label.text)
                return
        else:
            raise AssertionError(
                "Could not find a commit row with text %r" % expected_text)



from pathlib import Path
from setuptools import setup

MAIN_PACKAGE = 'heptapod_tests'
TESTS_PACKAGE = 'tests'
REQUIREMENTS_PATH = Path(__file__).parent / 'requirements.txt'

setup(
    name='heptapod-tests',
    version='0.9.0',
    author='octobus',
    author_email='contact@octobus.net',
    url='https://foss.heptapod.net/heptapod/heptapod-tests',
    long_description=Path('README.md').read_text(),
    long_description_content_type="text/markdown",
    description='Heptapod Functional Tests',
    keywords='hg mercurial git heptapod gitlab selenium REST GraphQL',
    packages=[MAIN_PACKAGE],
    license='GPL3+',
    install_requires=REQUIREMENTS_PATH.read_text().splitlines(),
)



# Copyright 2019-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 3 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import attr
import requests

from .access_levels import GroupAccess


class NameSpace:

    @property
    def url(self):
        return '/'.join((self.heptapod.url, self.full_path))


@attr.s
class UserNameSpace(NameSpace):
    user_name = attr.ib()
    heptapod = attr.ib()

    @property
    def full_path(self):
        return self.user_name

    def __eq__(self, other):
        return other.__class__ == self.__class__ and self.user == other.user


@attr.s
class Group(NameSpace):

    id = attr.ib()
    path = attr.ib()
    full_path = attr.ib()
    heptapod = attr.ib()
    owner_name = attr.ib(default=None)

    api_uri = '/api/v4/groups'

    @classmethod
    def api_create(cls, heptapod, group_path, user_name='root', parent=None):
        data = dict(name=group_path,
                    path=group_path)
        if parent is not None:
            data['parent_id'] = parent.id

        headers = {'Private-Token': heptapod.users[user_name].token}
        resp = requests.post(heptapod.url + cls.api_uri,
                             headers=headers,
                             data=data)
        assert resp.status_code == 201
        res = resp.json()
        return cls.api_retrieve(heptapod, group_id=res['id'])

    def __eq__(self, other):
        return self.id == other.id

    @classmethod
    def api_retrieve(cls, heptapod, group_id, owner_name=None):
        """Return a checked Group object for the given id.

        :owner_name: if specified, registered as :attr:`owner_name` on the
           returned object, and used for all API calls, including the check
           performed by this method.
        """
        grp = Group(heptapod=heptapod,
                    id=group_id,
                    full_path=None,
                    owner_name=owner_name,
                    path=None)
        resp = grp.api_get()
        assert resp.status_code == 200
        info = resp.json()
        grp.path = info['path']
        grp.full_path = info['full_path']
        return grp

    def private_token(self):
        """Return a token strong enough for all operations.

        Namely, using a token for an owner, if there's a known one or
        an Administrator token
        """
        user_name = self.owner_name if self.owner_name is not None else 'root'
        return {'Private-Token': self.heptapod.users[user_name].token}

    def api_get(self, subpath=''):
        return requests.get('/'.join((self.api_url, subpath)),
                            headers=self.private_token())

    def api_post(self, subpath='', **params):
        return requests.post('/'.join((self.api_url, subpath)),
                             headers=self.private_token(),
                             data=params)

    def api_put(self, subpath='', **params):
        return requests.put('/'.join((self.api_url, subpath)),
                            headers=self.private_token(),
                            data=params)

    @classmethod
    def api_search(cls, heptapod, group_name, user_name='root'):
        headers = {'Private-Token': heptapod.users[user_name].token}
        resp = requests.get(heptapod.url + cls.api_uri,
                            headers=headers,
                            params=dict(search=group_name))
        assert resp.status_code == 200
        data = resp.json()
        assert len(data) == 1
        group = data[0]
        return cls.api_retrieve(heptapod, group_id=group['id'])

    @property
    def api_url(self):
        return self.heptapod.url + self.api_uri + '/%d' % self.id

    def api_delete(self):
        resp = requests.delete(self.api_url, headers=self.private_token())
        assert resp.status_code in (202, 204)

    def grant_member_access(self, user, level):
        """Grant given user the given access level.

        It doesn't matter whether the user is already a member or not: this
        method abstracts over it.

        This method is idempotent.

        TODO factorize with Project (as surely is done in the Rails app)
        """
        assert level in GroupAccess

        user_id = user.id
        resp = self.api_get(subpath='members/%d' % user_id)
        if resp.status_code == 404:
            subpath = 'members'
            meth = self.api_post
        else:
            subpath = 'members/%d' % user_id
            meth = self.api_put

        resp = meth(subpath=subpath,
                    user_id=user_id,
                    access_level=int(level))
        assert resp.status_code < 400

    def custom_attribute_api_url(self, key):
        return '/'.join((self.api_url, 'custom_attributes', key))

    def api_set_custom_attribute(self, key, value, user=None):
        if user is None:
            user = self.heptapod.get_user('root')

        resp = requests.put(self.custom_attribute_api_url(key),
                            headers={'Private-Token': user.token},
                            data=dict(value=value))

        assert resp.status_code == 200
        return resp.json()

    def api_get_custom_attribute(self, key, check=True, user=None):
        if user is None:
            user = self.heptapod.get_user('root')

        resp = requests.get(self.custom_attribute_api_url(key),
                            headers={'Private-Token': user.token})

        if not check:
            return resp

        assert resp.status_code < 400
        return resp.json()['value']

    def fs_path(self):
        return '/'.join((self.heptapod.repositories_root, self.full_path))

    def put_hgrc(self, lines):
        """Replace group's server-side HGRC with given lines.

        The lines have to include LF, same as with `writelines()`.
        """
        fs_path = self.fs_path()
        self.heptapod.run_shell(('mkdir', '-p', fs_path), user='git')
        self.heptapod.put_file_lines(fs_path + '/hgrc', lines)

    def __enter__(self):
        return self

    def __exit__(self, *exc_args):
        self.api_delete()
        return False
